/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	GUI image viewport functionality										*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~										*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>

#include <types.h>
#include <memory.h>
#include <options.h>
#include <objects.h>
#include <distance.h>
#include <process.h>
#include "jobdefs.h"
#include "canvas.h"


/****************************************************************************/
/*							Prototypes										*/
/****************************************************************************/

static	void	JobCanvasExpose( GtkWidget *w, GdkEventExpose *e, JOB j );
static	void	JobBgndToRgb( JOB job );
static	void	JobFgndToRgb( JOB job );

static	void	JobObjectToMark( JOB job, DLINK *list, uword val );
static	void	JobBorderToMark( JOB job, DLINK *list, uword val );
static	void	JobDistToMark( JOB job );
static	void	JobOvlpToMark( JOB job );
static	void	JobTriaToMark( JOB job );
static	void	JobLineToMark( JOB job, RECT *line, int mask );
static	void	JobMarkToForeground( JOB job );

/****************************************************************************/
/*							Public functions								*/
/****************************************************************************/

/*
*	Creates the image canvas and possibly scrollbars
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

GtkWidget	*JobCanvasCreate( JOB job )
{
GtkWidget		*drw;
GtkWidget		*win;
GtkWidget		*scr;
GtkAdjustment	*adj;

	// Create the GDK canvas
	
	drw = gtk_drawing_area_new();
	gtk_drawing_area_size( GTK_DRAWING_AREA( drw ), job->xs, job->ys );
	gtk_widget_show( drw );
	
	// Create a viewport window and fix its size, set scrollbars to be
	// always on
	
	win = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW( win ), drw );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( win ),
									GTK_POLICY_ALWAYS, 
									GTK_POLICY_ALWAYS );
	gtk_widget_show( win );
	
	// Set up the expose event handler
		
	gtk_signal_connect( GTK_OBJECT( drw ), "expose_event",
						GTK_SIGNAL_FUNC( JobCanvasExpose ), job );
	
	// Enable mouse events, including hints. It is needed for containment.
	
	gtk_widget_set_events( drw, GDK_EXPOSURE_MASK | 
								GDK_LEAVE_NOTIFY_MASK |
								GDK_BUTTON_PRESS_MASK | 
								GDK_POINTER_MOTION_MASK |
								GDK_POINTER_MOTION_HINT_MASK );
	
	// Store the canvas reference in the job structure, create the
	// RGB image and return the viewport widget
	
	job->rawrgb = Malloc( job->ys * job->xs * 3 );
	job->canvas = drw;
	job->line	= NULL;
	job->cxs	= job->xs;
	job->cys	= job->ys;
	JobCanvasRefresh( job );
	return( win );
}

/*
*	Sets the magnification
*	~~~~~~~~~~~~~~~~~~~~~~
*/

void	JobCanvasMagnify( JOB job, int new )
{
	if ( job->line ) {
		
		Free( job->line );
		job->line = NULL;
	}
	
	if ( new != 1 ) job->line = Malloc( (job->xs+1) * 3 * new );
	job->magnif = new;
	gtk_widget_hide( job->canvas );
	gtk_drawing_area_size( GTK_DRAWING_AREA( job->canvas ), 
						   job->xs * new, job->ys * new );
	gtk_widget_show( job->canvas );
	JobCanvasRefresh( job );
}

/*
*	Destroys the canvas and the viewport
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	We destroy the canvas itself, that will notify the
*	container window that there's nothing to hold any more.
*	We then can destroy the raw RGB image.
*/

void	JobCanvasDestroy( JOB job )
{
	if ( job->canvas ) gtk_widget_destroy( job->canvas );
	if ( job->rawrgb ) Free( job->rawrgb );
	if ( job->line   ) Free( job->line );
	if ( job->fgnd   ) ImageDestroy( job->fgnd );
	job->canvas = NULL;
	job->rawrgb = NULL;
	job->line	= NULL;
	job->fgnd	= NULL;
}

/*
*	Creates the image that is being displayed
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	JobCanvasRefresh( JOB job )
{
GdkRectangle	rect;
int				x, y;

	y = job->ys;
	x = job->xs;
	
	memset( job->rawrgb, 0, 3 * x * y );
	JobBgndToRgb( job );
	JobFgndToRgb( job );
	
	rect.x = 0;
	rect.y = 0;
	rect.width  = x * job->magnif;
	rect.height = y * job->magnif;
	gtk_widget_draw( job->canvas, &rect );
}

/*
*	Calculates the mark according to current objects
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	It also recalculates the foreground
*/

void	JobCanvasObjectMap( JOB job )
{
int		y;

	for ( y = 0 ; y < job->ys ; y++ ) {
		
		memset( job->mark[ y ], 0, job->xs * sizeof( uword ) );
	}
		
	// Draw the objects
	
	JobObjectToMark( job, &job->act, MARK_ACT_INT );
	JobObjectToMark( job, &job->ref, MARK_REF_INT );
	JobObjectToMark( job, &job->con, MARK_CON_INT );
			
	// Draw the perimeters
			
	JobBorderToMark( job, &job->act, MARK_ACT_OUT );
	JobBorderToMark( job, &job->ref, MARK_REF_OUT );
	JobBorderToMark( job, &job->con, MARK_CON_OUT );

	// Draw the distances
	
	JobDistToMark( job );
	
	// Draw the overlaps
	
	JobOvlpToMark( job );
	
	// Draw the triangles
	
	JobTriaToMark( job );
}

/*
*	Creates the foreground image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	JobCanvasForeground( JOB job )
{
int		y;

	if ( ! job->opt_ovlp && ! job->opt_objs && 
		 ! job->opt_dist && ! job->opt_conf && 
		 ! job->opt_tria ) {
	
		// No foreground at all
		
		if ( job->fgnd ) {
		
			ImageDestroy( job->fgnd );
			job->fgnd = NULL;
		}
	}
	else {
		
		if ( ! job->fgnd ) {
	
			job->fgnd = ImageCreate( job->ys, job->xs, 3 );
		}
		else {
	
			ImageClear( job->fgnd, R, 0 );
			ImageClear( job->fgnd, G, 0 );
			ImageClear( job->fgnd, B, 0 );
		}
		
		JobMarkToForeground( job );
	}
}

/****************************************************************************/
/*							Local functions									*/
/****************************************************************************/

/*
*	Copies the background image to the RGB display data
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobBgndToRgb( JOB job )
{
ubyte	*rgb;
int		x, y, l;
int		r, g, b;
int		ra, ga, ba;
int		rr, gr, br;

	if ( ! job->bgnd || ! *job->bgnd ) return;
	
	if ( (*job->bgnd)->l == 3 ) {
	
		for ( l = 0 ; l < 3 ; l++ ) {
		
			for ( rgb = job->rawrgb + l, y = 0 ; y < job->ys ; y++ ) {
	
				for ( x = 0 ; x < job->xs ; x++ ) {
		
					*rgb = (*job->bgnd)->im[ l ][ y ][ x ];
					rgb += 3;
				}
			}
		}
	}
	else {
			
		OptionValueGet( optColRefObj, &l );
		rr = ( l >> 16 ) & 255;
		gr = ( l >>  8 ) & 255;
		br = ( l >>  0 ) & 255;
		l = MAX( rr, MAX( gr, br ) );
		rr = rr * 255 / l;
		gr = gr * 255 / l;
		br = br * 255 / l;
		
		OptionValueGet( optColActObj, &l );
		ra = ( l >> 16 ) & 255;
		ga = ( l >>  8 ) & 255;
		ba = ( l >>  0 ) & 255;
		l = MAX( ra, MAX( ga, ba ) );
		ra = ra * 255 / l;
		ga = ga * 255 / l;
		ba = ba * 255 / l;
			
		rgb = job->rawrgb;
				
		for ( rgb = job->rawrgb, y = 0 ; y < job->ys ; y++ ) {
	
			for ( x = 0 ; x < job->xs ; x++ ) {
			
				r  = ( (*job->bgnd)->im[ R ][ y ][ x ] * ra ) >> 8;
				g  = ( (*job->bgnd)->im[ R ][ y ][ x ] * ga ) >> 8;
				b  = ( (*job->bgnd)->im[ R ][ y ][ x ] * ba ) >> 8;
					
				r += ( (*job->bgnd)->im[ G ][ y ][ x ] * rr ) >> 8;
				g += ( (*job->bgnd)->im[ G ][ y ][ x ] * gr ) >> 8;
				b += ( (*job->bgnd)->im[ G ][ y ][ x ] * br ) >> 8;
				
				*rgb++ = r < 255 ? r : 255;
				*rgb++ = g < 255 ? g : 255;
				*rgb++ = b < 255 ? b : 255;
			}
		}
	}
}

/*
*	Copies the foreground image to the RGB image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	If a pixel is (0,0,0), then it will not be modified in the
*	RGB. Otherwise, it will be changed.
*	The foreground image is always a 3-layer image.
*/

static	void	JobFgndToRgb( JOB job )
{
ubyte	*rgb;
int		x, y;
ubyte	r, g, b;

	if ( ! job->fgnd ) return;
	
	for ( rgb = job->rawrgb, y = 0 ; y < job->ys ; y++ ) {

		for ( x = 0 ; x < job->xs ; x++ ) {
		
			r = job->fgnd->im[ R ][ y ][ x ];
			g = job->fgnd->im[ G ][ y ][ x ];
			b = job->fgnd->im[ B ][ y ][ x ];
	
			if ( r | g | b ) {
			
				*rgb++ = r;
				*rgb++ = g;
				*rgb++ = b;
			}
			else {
			
				rgb += 3;
			}
		}
	}
}

/*
*	Fills the mark with all the objects
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobObjectToMark( JOB job, DLINK *list, uword val )
{
OBJECT	p;
PIXEL	m;
int		x, y;

	for ( p = DlinkFirst( list ) ; p ; p = DlinkNext( p ) ) {
				
		for ( m = ObjectFirstPixel( p, FALSE, &y, &x ) ; 
			  m ; m = ObjectNextPixel( p, m, &y, &x ) ) {
		
			if ( ( x + y ) % 6 == 0 ) {
			
				job->mark[ y ][ x ] |= val;
			}
		}
	}
}

/*
*	Fills the mark with all the object boundaries
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobBorderToMark( JOB job, DLINK *list, uword val )
{
OBJECT	p;
PIXEL	m;
int		x, y;

	for ( p = DlinkFirst( list ) ; p ; p = DlinkNext( p ) ) {
				
		for ( m = ObjectFirstPixel( p, TRUE, &y, &x ) ; 
			  m ; m = ObjectNextPixel( p, m, &y, &x ) ) {
		
			job->mark[ y ][ x ] |= val;
		}
	}
}

/*
*	Fills the mark with the distances
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	The distance endpoints are marked with a 3x3 dot and a line
*	is drawn using the Breshenham algorithm
*/

static	void	JobDistToMark( JOB job )
{
DIST	*dp;

	for ( dp = DlinkFirst( &job->dist ) ; dp ; dp = DlinkNext( dp ) ) {
	
		JobLineToMark( job, &dp->pair, MARK_DISTANCE );
	}
}
	
/*
*	Fills the mark with the overlaps
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	The distance endpoints are marked with a 3x3 dot and a line
*	is drawn using the Breshenham algorithm
*/

static	void	JobOvlpToMark( JOB job )
{
OVLP	*ovl;
PIXEL	pix;
int		x, y;

	for ( ovl = DlinkFirst( &job->ovlp ) ; ovl ; ovl = DlinkNext( ovl ) ) {
	
		for ( pix = DlinkFirst( &ovl->pixs ) ; pix ; pix = DlinkNext( pix ) ) {
		
			ObjectPixelCoords( pix, &y, &x );
			job->mark[ y ][ x ] |= MARK_OVERLAP;
		}
	}
}

/*
*	Draws the triangles to the mark
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobTriaToMark( JOB job )
{
DIST	*tp;

	for ( tp = DlinkFirst( &job->ltri ) ; tp ; tp = DlinkNext( tp ) ) {
	
		JobLineToMark( job, &tp->pair, MARK_TRIANGLE );
	}	
}

/*
*	Creates the foreground image from the mark
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobMarkToForeground( JOB job )
{
int		y, x;
int		r, g, b;
int		val;
int		cact, cref, ccon, cdis;

	OptionValueGet( optColRefObj,	&cref );
	OptionValueGet( optColActObj,	&cact );
	OptionValueGet( optColDistance, &cdis );
	OptionValueGet( optColConfine,	&ccon );
	
	if ( ! job->mask ) return;
	
	for ( y = 0 ; y < job->ys ; y++ ) {
	
		for ( x = 0 ; x < job->xs ; x++ ) {
		
			if ( ! ( val = job->mark[ y ][ x ] & job->mask ) ) continue;
			
			r = g = b = 0;
			
			if ( val & ( MARK_ACT_INT | MARK_ACT_OUT ) ) {			
			
				r += ( cact >> 16 ) & 255;
				g += ( cact >>  8 ) & 255;
				b += ( cact >>  0 ) & 255;
			}
			
			if ( val & ( MARK_REF_INT | MARK_REF_OUT ) ) {			
			
				r += ( cref >> 16 ) & 255;
				g += ( cref >>  8 ) & 255;
				b += ( cref >>  0 ) & 255;
			}
			
			if ( val & ( MARK_CON_INT | MARK_CON_OUT ) ) {			
			
				r += ( ccon >> 16 ) & 255;
				g += ( ccon >>  8 ) & 255;
				b += ( ccon >>  0 ) & 255;
			}
			
			if ( val & ( MARK_DISTANCE | MARK_OVERLAP | MARK_TRIANGLE ) ) {
			
				r += ( cdis >> 16 ) & 255;
				g += ( cdis >>  8 ) & 255;
				b += ( cdis >>  0 ) & 255;
			}
			
			if ( r > 255 ) r = 255;
			if ( g > 255 ) g = 255;
			if ( b > 255 ) b = 255;
			
			job->fgnd->im[ R ][ y ][ x ] = r;
			job->fgnd->im[ G ][ y ][ x ] = g;
			job->fgnd->im[ B ][ y ][ x ] = b;
		}
	}
}

/*
*	Handles the expose event for the canvas
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobCanvasExpose( GtkWidget *g, GdkEventExpose *e, JOB j )
{
int		rsy, rsx, ycn, xcn;
char	*rgbp, *rptr, *vptr;
int		vcnt, i;

	// Make sure that the values are making sense
	
	if ( e->area.x >= j->cxs * j->magnif ) return;
	if ( e->area.y >= j->cys * j->magnif ) return;
		
	if ( e->area.x + e->area.width >= j->cxs * j->magnif ) 
			
		e->area.width = j->cxs * j->magnif - e->area.x;
		
	if ( e->area.y + e->area.height >= j->cys * j->magnif ) 
			
		e->area.height = j->cys * j->magnif - e->area.y;
		
	// If the magnification is 1, we don't have to do anything fancy,
	// leave the work to GTK
	
	if ( j->magnif == 1 ) {
	
		rgbp = j->rawrgb + ( e->area.y * j->cxs + e->area.x ) * 3;
		
		gdk_draw_rgb_image( g->window, 
							g->style->fg_gc[ GTK_STATE_NORMAL ],
							e->area.x, 
							e->area.y, 
							e->area.width, 
							e->area.height,
							GDK_RGB_DITHER_MAX,
							rgbp,
							j->cxs * 3 );
		return;
	}
	
	// We have to do some work. Calculate our own start 
	// coordinates from the requested virtual ones
	
	rsy = e->area.y / j->magnif;
	rsx = e->area.x / j->magnif;
	
	// Calculate the y and x offsets between the requested virtual
	// start coordinates and the ones that our real coordinates map
		
	ycn = e->area.y - rsy * j->magnif;
	xcn = e->area.x - rsx * j->magnif;

	// Limit ycn if the requested heighth is less than the first
	// batch we would schedule
	
	if ( j->magnif - ycn > e->area.height ) {
	
		ycn = j->magnif - e->area.height;
	}
	
	// Work out the starting index into the RGB array
	
	rgbp = j->rawrgb + ( rsy * j->cxs + rsx ) * 3;
	
	// Start generating the lines and send them to the screen
	
	while ( e->area.height > 0 ) {
		
		// Generate the magnified versions of the pixels from our image
		
		rptr = rgbp;
		vptr = j->line;
		vcnt = e->area.width + j->magnif - 1; 
		
		while ( vcnt > 0 ) {
		
			for ( i = 0 ; i < j->magnif ; i++ ) {
			
				*vptr++ = rptr[ 0 ];
				*vptr++ = rptr[ 1 ];
				*vptr++ = rptr[ 2 ];
			}
			
			rptr += 3;
			vcnt -= j->magnif;
		}
		
		// OK, now we have enough pixels to send it to the canvas
		
		gdk_draw_rgb_image( g->window, 
							g->style->fg_gc[ GTK_STATE_NORMAL ],
							e->area.x,
							e->area.y,
							e->area.width,
							j->magnif - ycn,	
							GDK_RGB_DITHER_MAX,
							j->line + 3 * xcn,
							0 );
							
		// We finished with this line, update the remaining area
		
		e->area.y	   += j->magnif - ycn;
		e->area.height -= j->magnif - ycn;
		
		// Work out how many to schedule in the next round
				
		ycn = e->area.height > j->magnif ? 0 : j->magnif - e->area.height;
		
		// Update source pointer
		
		rgbp += 3 * j->cxs;
	}
}

/*
*	Draws a line with the given value onto mark
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobLineToMark( JOB job, RECT *line, int mask )
{
int		x1, y1, x2, y2;
int		dx, dy, sx, sy;
int		i, d;
	
	x1 = line->tlx;
	y1 = line->tly;
	x2 = line->brx;
	y2 = line->bry;
		
	for ( dy = -1 ; dy <= 1 ; dy++ ) {
		
		for ( dx = -1 ; dx <= 1 ; dx++ ) {
			
			sx = x1 + dx;
			sy = y1 + dy;
				
			if ( sx >= 0 && sx < job->orig->x &&
				 sy >= 0 && sy < job->orig->y ) {
					 
				job->mark[ sy ][ sx ] |= mask;
			}
				
			sx = x2 + dx;
			sy = y2 + dy;
			
			if ( sx >= 0 && sx < job->orig->x &&
				 sy >= 0 && sy < job->orig->y ) {
				 
				job->mark[ sy ][ sx ] |= mask;
			}
		}
	}
	
	dx = ABS( x1 - x2 );
	dy = ABS( y1 - y2 );
	
	sx = ( x2 > x1 ) ? 1 : -1;
	sy = ( y2 > y1 ) ? 1 : -1;

	if ( dx >= dy ) {
	
		d = dx / 2;
		
		for ( i = 0 ; i < dx ; i++ ) {
		
			job->mark[ y1 ][ x1 ] |= mask;
			
			d += dy;
		
			if ( d >= dx ) {
		
				y1 += sy;
				d  -= dx;
			}
		
			x1 += sx;
		}
	}
	else {

		d = dy / 2;
	
		for ( i = 0 ; i < dy ; i++ ) {
	
			job->mark[ y1 ][ x1 ] |= mask;
		
			d += dx;
		
			if ( d >= dy ) {
		
				x1 += sx;
				d -= dy;
			}
		
			y1 += sy;
		}
	}

	job->mark[ y2 ][ x2 ] |= mask;
}

