/****************************************************************************/
/* Copyright 2004, © Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	GUI image processing steps												*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~												*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltán Kócsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <math.h>

#include <types.h>
#include <job.h>
#include <objects.h>
#include <process.h>
#include <options.h>
#include <progress.h>
#include <distance.h>
#include <randomlist.h>
#include <stats.h>
#include <memory.h>
#include "jobdefs.h"
#include "process.h"
#include "jobgui.h"

#define	STAT_TESTS	100

/****************************************************************************/
/****************************************************************************/

static	void	JobSeparateDestroy( JOB job );
static	void	JobEqualiseDestroy( JOB job );
static	void	JobExtractDestroy( JOB job );
static	void	JobDistanceDestroy( JOB job );
static	void	JobStatisticsDestroy( JOB job );
static	DATA	*JobStatisticsToData( DLINK *dist, DLINK *ovlp );

/****************************************************************************/
/****************************************************************************/

/*
*	Releases all process related resources of an image 
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	JobRelease( JOB job )
{
	JobSeparateDestroy( job );
}

/*
*	Preprocesses the image for a given job
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Returns 1 if succeeded, 0 if there was nothing to do and -1
*	if it was aborted.
*/

int		JobSeparate( JOB job )
{
IMAGE	*temp;
double	frad;
int		csep, acol, rcol;

	csep = job->new.csep;
	acol = job->new.acol;
	rcol = job->new.rcol;
	frad = job->new.frad;
	
	// If all options are the same as before and we already have
	// a preprocessed image, we don't have to work
	
	if ( job->filt && 
		 job->opt.rcol == rcol && job->opt.acol == acol &&
		 job->opt.csep == csep && job->opt.frad == frad ) {
		 
		return( FALSE );
	}
	
	JobSeparateDestroy( job );
	
	// Update parameters
	
	job->opt.csep = csep;
	job->opt.frad = frad;
	job->opt.acol = acol;
	job->opt.rcol = rcol;
	
	// Process the images
	
	if ( ! ( temp = ColourSepImage( job->orig, csep, rcol, acol ) ) ) {
	
		// Colour separation was aborted
		
		return( ABORT );
	}
	
	job->filt = FilterImage( temp, frad );
	ImageDestroy( temp );
	
	if ( ! job->filt ) {
	
		// Filtering was aborted
		
		return( ABORT );
	}
	
	return( TRUE );
}

/*
*	Equalises and gamma corrects an image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

int		JobEqualise( JOB job )
{
int		flag;
int		eqal;
double	gamm;

	if ( JobSeparate( job ) == ABORT ) {
	
		// There was separation to be done and it was aborted,
		// report the sad news and leave
		
		return( ABORT );
	}
	
	// We do not have to test for TRUE or FALSE, because if JobSeparate 
	// had to do anything, it would have destroyed our image, if it had 
	// existed before.
		
	eqal = job->new.eqal;
	gamm = job->new.gamm;
	
	if ( job->gamm && eqal == job->opt.eqal && gamm == job->opt.gamm ) {
	
		// Nothing has changed, nothing to do
		
		return( FALSE );
	}
	
	// We have to process the image - get rid of all further results
	
	JobEqualiseDestroy( job );
	
	// Update parameters
		
	job->opt.gamm = gamm;
	job->opt.eqal = eqal;
	
	if ( ! ( job->gamm = GammaImage( job->filt, eqal, gamm ) ) ) {
	
		// It was aborted
		
		return( ABORT );
	}
	
	return( TRUE );
}

/*
*	Extracts the objects from the preprocessed image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Returns if it actually did something
*/

int		JobExtract( JOB job )
{
int		meth, seed, acpt, mins, fill, excl;
int		flag;
int		y, x;

	flag = FALSE;
	
	if ( JobEqualise( job ) == ABORT ) {
	
		return( ABORT );
	}
	
	y = job->gamm->y;
	x = job->gamm->x;
		
	meth = job->new.act_meth;
	seed = job->new.act_seed;
	acpt = job->new.act_acpt;
	mins = job->new.act_mins;
	fill = job->new.act_fill;
	excl = job->new.act_excl;
	
	if ( ! DlinkFirst( &job->act ) || 
		 meth != job->opt.act_meth || excl != job->opt.act_excl ||
		 seed != job->opt.act_seed || acpt != job->opt.act_acpt ||
		 mins != job->opt.act_mins || fill != job->opt.act_fill ||
		 job->opt.dis_meth != job->new.dis_meth ||
		 job->opt.pair_mic != job->new.pair_mic ||
		 job->opt.pair_pix != job->new.pair_pix ) {
	
		// We need to extract the active layer
		// First, release whatever we used to have
		
		ObjectDestroyList( &job->act );
		JobDistanceDestroy( job );
		
		// Update options
		
		job->opt.act_meth = meth;
		job->opt.act_seed = seed;
		job->opt.act_acpt = acpt;
		job->opt.act_mins = mins;
		job->opt.act_fill = fill;
		job->opt.act_excl = excl;
		job->opt.pair_pix = job->new.pair_pix;
		job->opt.pair_mic = job->new.pair_mic;
		job->opt.dis_meth = job->new.dis_meth;
		
		// Extarct objects
		
		if ( ObjectSearch( &job->act, job->gamm, R, &job->con,
						   "Extracting active objects",
						   meth, seed, acpt, mins ) == ABORT ) {
							 
			JobExtractDestroy( job );
			return( ABORT );
		}
		
		// Postprocess object list
		
		job->actlen = ObjectFixup( &job->act, y, x, 
								   "Postprocessing active objects", 
								   mins, fill, excl );
								   
		if ( job->actlen == ABORT ) {
		
			JobExtractDestroy( job );
			return( ABORT );
		}
		
		// For the active objects we also do the triangulation

		switch ( job->opt.dis_meth ) {
			
			case OPT_SELF_TRIA:
		
				job->ntri = Delaunay( &job->act, &job->ltri );
				break;
			
			case OPT_SELF_TREE:
			
				job->ntri = Dijkstra( &job->act, &job->ltri );
				break;
				
			case OPT_SELF_NEIG:
			
				job->ntri = Neighbour( &job->act, &job->ltri );
				break;
				
			case OPT_SELF_RATS:
			
				job->ntri = Ratsnest( &job->act, &job->ltri );
				break;
		}
		
		// Make note that we've done something
		
		flag = TRUE;
	}
	
	meth = job->new.ref_meth;
	seed = job->new.ref_seed;
	acpt = job->new.ref_acpt;
	mins = job->new.ref_mins;
	fill = job->new.ref_fill;
	excl = job->new.ref_excl;

	if ( ! DlinkFirst( &job->ref ) || 
		 meth != job->opt.ref_meth || excl != job->opt.ref_excl ||
		 seed != job->opt.ref_seed || acpt != job->opt.ref_acpt ||
		 mins != job->opt.ref_mins || fill != job->opt.ref_fill ) {
	
		// We need to extract the reference layer
		// Discard old stuff
		
		ObjectDestroyList( &job->ref );
		JobDistanceDestroy( job );
		
		// Update options
		
		job->opt.ref_meth = meth;
		job->opt.ref_seed = seed;
		job->opt.ref_acpt = acpt;
		job->opt.ref_mins = mins;
		job->opt.ref_fill = fill;
		job->opt.ref_excl = excl;
		
		if ( ObjectSearch( &job->ref, job->gamm, G, &job->con,
						   "Extracting reference objects",
					       meth, seed, acpt, mins ) == ABORT ) {
							 
			JobExtractDestroy( job );
			return( ABORT );
		}
		
		// Postprocess object list
		
		job->reflen = ObjectFixup( &job->ref, y, x, 
								   "Postprocessing reference objects", 
								   mins, fill, excl );
								   
		if ( job->reflen == ABORT ) {
		
			JobExtractDestroy( job );
			return( ABORT );
		}
		
		flag = TRUE;
	}
	
	return( flag );
}

/*
*	Maps the distances of the reference and active objects
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Returns if it actually did something
*/

int		JobDistance( JOB job )
{
int		maxd;
int		maxa;
int		maxr;

	if ( JobExtract( job ) == ABORT ) {
	
		return( ABORT );
	}
	
	maxd = job->new.pair_dis;
	maxa = job->new.pair_act;
	maxr = job->new.pair_ref;
	
	if ( DlinkFirst( &job->dist )  && DlinkFirst( &job->ovlp )  &&
		 maxd == job->opt.pair_dis && maxa == job->opt.pair_act && 
		 maxr == job->opt.pair_ref ) {

		return( FALSE );
	}
	
	// Update parameters
	
	job->opt.pair_dis = maxd;
	job->opt.pair_act = maxa;
	job->opt.pair_ref = maxr;
	
	JobDistanceDestroy( job );
	
	if ( DistMapLists( &job->dist, &job->ovlp, &job->ref, &job->act, 
					   maxd, maxr, maxa, TRUE ) == ABORT ) {
	
		return( ABORT );
	}
	
	return( TRUE );
}

/*
*	Processes the statistics for an image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

int		JobStats( JOB job )
{
DIST	*dp;
RECT	bound;
DLINK	list;
DLINK	dist;
DLINK	ovlp;
int		i, n;
jmp_buf	state;
DLINK	empty;
int		do_dist;
int		do_self;
int		self_meth;

	JobDistance( job );
	
	if ( job->stat && 
		 job->new.stat_num == job->opt.stat_num &&
		 job->new.stat_act == job->opt.stat_act &&
		 job->new.stat_ref == job->opt.stat_ref ) {
	
		return( FALSE );
	}
	
	JobStatisticsDestroy( job );
	
	DlinkInit( &list );
	DlinkInit( &dist );
	DlinkInit( &ovlp );
	DlinkInit( &empty );
	
	bound.tly = 0;
	bound.tlx = 0;
	bound.bry = job->ys;
	bound.brx = job->xs;
	
	if ( setjmp( state ) ) {
	
		// Aborted
		
		if ( job->stat ) StatDestroy( (SINFO) job->stat );
		if ( job->tris ) StatDestroy( (SINFO) job->tris );
		job->stat = NULL;
		job->tris = NULL;
		return( ABORT );
	}
	
	
	if ( job->new.stat_ref ) {
	
		job->stat = (void *) StatPrepare( &job->dist, &job->ovlp );
		DlinkAddHead( &job->dsta, 
					  JobStatisticsToData( &job->dist, &job->ovlp ) );
	}
	
	if ( job->new.stat_act ) {
	
		job->tris = (void *) StatPrepare( &job->ltri, &empty );
		DlinkAddHead( &job->tsta, 
					  JobStatisticsToData( &job->ltri, &empty ) );
	}
	
	n = job->new.stat_num;
	
	ProgressInit( "Generating random images", state );
	
	for ( i = 0 ; i < n ; i++ ) {
		
		ProgressUpdate( i, n );
		
		// Create a random distribution of objects
		
		if ( job->new.stat_ref || job->new.stat_act )
			
			RandomListCreate( &bound, &list, &job->act, &job->con );
		
		if ( job->new.stat_ref ) {
		
			// Calculate the distances and the overlaps
		
			DistMapLists( &dist, &ovlp, &job->ref, &list, job->opt.pair_dis, 
						  job->opt.pair_ref, job->opt.pair_act, FALSE );
		
			// Process the distances. The dist and ovlp lists will
			// be retained by the statistics routine, they MUST NOT
			// be freed (they will be empty at return, actually).
		
			DlinkAddTail( &job->dsta, 
						  JobStatisticsToData( &dist, &ovlp ) );
					  
			StatRound( (void *) job->stat, &dist, &ovlp );
		}
		
		// Work out the triange segment things
		
		if ( job->new.stat_act ) {
		
			switch ( job->opt.dis_meth ) {
			
				case OPT_SELF_TRIA:
		
					Delaunay( &list, &dist );
					break;
			
				case OPT_SELF_TREE:
					
					Dijkstra( &list, &dist );
					break;
					
				case OPT_SELF_NEIG:
					
					Neighbour( &list, &dist );
					break;
					
				case OPT_SELF_RATS:
					
					Ratsnest( &list, &dist );
					break;
			}
			
			DlinkAddTail( &job->tsta, 
						  JobStatisticsToData( &dist, &empty ) );
					  
			StatRound( (void *) job->tris, &dist, &empty );
		}
		
		// Release temporary stuff
		
		RandomListFree( &list );
	}
	
	ProgressDone();
	if ( job->new.stat_ref ) job->stat = StatFinal( (void *) job->stat );
	if ( job->new.stat_act ) job->tris = StatFinal( (void *) job->tris );
	job->opt.stat_num = n;
	job->opt.stat_ref = job->new.stat_ref;
	job->opt.stat_act = job->new.stat_act;
	return( TRUE );
}

/*
*	Invalidates anything that is affected by confinement changes
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	JobConfinementChanged( JOB job )
{
	// Delete everything including and after the feature extraction
	
	JobExtractDestroy( job );
}

/****************************************************************************/
/****************************************************************************/

/*
*	Destroys the data associated with colour separation
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobSeparateDestroy( JOB job )
{
	JobEqualiseDestroy( job );
	
	if ( job->filt ) {
	
		ImageDestroy( job->filt );
		job->gamm = NULL;
	}
}

/*
*	Destroys the data associated with gamma correction
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobEqualiseDestroy( JOB job )
{
	JobExtractDestroy( job );
	
	if ( job->gamm ) {
	
		ImageDestroy( job->gamm );
		job->gamm = NULL;
	}
}

/*
*	Destroys the data associated with extraction
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobExtractDestroy( JOB job )
{
void	*d;

	JobDistanceDestroy( job );
	ObjectDestroyList( &job->act );
	ObjectDestroyList( &job->ref );
	while ( ( d = DlinkRemoveFirst( &job->ltri ) ) ) Free( d );
	job->actlen = 0;
	job->reflen = 0;
	job->ntri	= 0;	
}

/*
*	Destroys the data associated with distance lists
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobDistanceDestroy( JOB job )
{
void	*d;

	JobStatisticsDestroy( job );
	while ( ( d = DlinkRemoveFirst( &job->dist ) ) ) Free( d );
	while ( ( d = DlinkRemoveFirst( &job->ovlp ) ) ) Free( d );
}

/*
*	Destroys the data associated with statistics
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	JobStatisticsDestroy( JOB job )
{
DATA	*d;

	while ( ( d = DlinkRemoveFirst( &job->dsta ) ) ) Free( d );
	while ( ( d = DlinkRemoveFirst( &job->tsta ) ) ) Free( d );
	if ( job->stat ) Free( job->stat );
	if ( job->tris ) Free( job->tris );
	job->stat = NULL;
	job->tris = NULL;
}

/*
*	Creates a data entry from a distance, overlap list pair
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	DATA	*JobStatisticsToData( DLINK *dist, DLINK *ovlp )
{
DIST	*dp;
OVLP	*op;
DATA	*rp;
int		n;

	for ( n = 0, dp = DlinkFirst( dist ) ; dp ; dp = DlinkNext( dp ), n++ );
	for ( op = DlinkFirst( ovlp ) ; op ; op = DlinkNext( op ), n++ );

	rp = Malloc( sizeof( DATA ) + sizeof( double ) * n );
	rp->numd = n;
	
	for ( n = 0, dp = DlinkFirst( dist ) ; dp ; dp = DlinkNext( dp ), n++ ) {
	
		rp->data[ n ] = sqrt( (double) dp->dist );
	}
	
	for ( op = DlinkFirst( ovlp ) ; op ; op = DlinkNext( op ), n++ ){
	
		rp->data[ n ] = 1 - sqrt( (double) op->nump );
	}
	
	return( rp );		
}
