/****************************************************************************/
/* Copyright 2004, � Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Feature processing related things										*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~										*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zolt�n K�csi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#ifndef __PROCESS_H__
#define __PROCESS_H__

#include <types.h>
#include <image.h>
#include <dlinklist.h>
#include <objects.h>

/*
*	Picks the colours according to the selector value
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*
*		old:	Original image
*		method:	Colour separation method
*		ric:	Reference feature colour in an int (3 bytes:00rrggbb)
*		aic:	Active feature colour in an int (3 bytes:00rrggbb)
*	
*	Return value
*
*		A pointer to the new image or NULL if could not finish
*
*	Receives an original image and returns a new one, the same size,
*	but without the B layer. The resulting image will contain the
*	active objects on the R layer and the reference on the G.
*	If the opration was aborted, returns NULL.
*/

IMAGE	*ColourSepImage( IMAGE *old, int method, int ric, int aic );

/*
*	Gamma corrects and equalises an image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*
*		old:	Original image
*		equal:	Level equalisation is required
*		gamm:	Gamma value
*	
*	Return value
*
*		A pointer to the new image or NULL if could not finish
*
*	Receives a 2-layer image. For both layers, scale the highest values
*	to 255 and run a gamma correction on all pixels with the given
*	gamma value. Return the resulting image or NULL, if the processing
*	was aborted.
*/

IMAGE	*GammaImage( IMAGE *image, int equal, double gamm );

/*
*	Filters the R and G layers on the image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*
*		old:		Original image
*		radius:		The radius of the filter kernel
*	
*	Return value
*
*		A pointer to the new image or NULL if could not finish
*
*	Receives a 2-layer image. For both layers, applies a convolution
*	filter with a raised cosine kernel of the given radius.
*	Returns the resulting image or NULL, if the processing was aborted.
*/

IMAGE	*FilterImage( IMAGE *image, double radius );

/*
*	Finds the objects on an image layer, using the selected algorithm
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*
*		list		A DLINK onto which the feature objects will be placed
*		image		The image on which the search should be done
*		layer		The layer of the image
*		conf		A list of objects representing excluded areas
*		prog		Title for the porgress bar
*		method		Extraction method
*		seed		Minimum seed value for feature growing
*		acpt		Minimum value to consider a pixel
*		size		Minimum feature size (for blobsearch only)
*	
*	Return value
*
*		TRUE if succeeded, FALSE if it was aborted.
*
*	It does not do a list trimming, just finds anything that is
*	acceptable by the search algorithm. Places the objects onto
*	the DLINK passed to it. The DLINK will be empty if aborted.
*/

int		ObjectSearch( DLINK *list, IMAGE *image, int layer, DLINK *conf,
					  char *prog, int method, int seed, int acpt, int size );

/*
*	Turns a bitmap into an object list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*
*	Parameters
*
*		list		A DLINK onto which the objects will be placed
*		bitmap		An array of chars, representing the bitmapped image
*		ys, xs		Size of the array
*
*	Return value
*		-
*
*	Each pixel is represented as a byte on the bitmap. If the byte is
*	non-zero, the pixel will be part of an object. The objects represent
*	continuous marked areas of the bitmap.
*/

void	ObjectMapped( DLINK *list, ubyte **bitmap, int ys, int xs );

/*
*	Trimms all objects on a list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		objs		A pointer to a DLINK containing the objects to trim
*		y, x		The size of the image
*		txt			Text for the progress bar
*		size		Minimal size (in pixels) for an object
*		fill		Flag indicating if holes should be filled
*		excl		Flag indicating if boundary objects should be discarded
*
*	Return value
*		The length of the final queue or ABORT if it was aborted
*/

int		ObjectFixup( DLINK *objs, int y, int x, char *txt, 
					 int size, int fill, int excl );

/*
*	Calculates the perimeters and the spatial trees for a list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		objs		A pointer to a DLINK containing the objects to trim
*		y, x		The size of the image
*
*	Return value
*		The length of the final queue
*
*	This function does not change the objects whatsoever. 
*	It merely calculates the borders and the creates distance division trees.
*/

int		ObjectQuickfix( DLINK *objs, int y, int x );

/*
*	From a list of objects creates the triangulation data
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		list		A list of objects
*		tria		A list of segments for the triangles (distance objects)
*
*	Return value:
*		The number of segments
*/

int		Delaunay( DLINK *list, DLINK *tria );

/*
*	Selects the lines that form the minimum cost spanning tree
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		list		A list of objects
*		tria		A list of segments for the tree edges (distance objects)
*
*	Return value:
*		The number of segments
*/

int		Dijkstra( DLINK *list, DLINK *tria );

/*
*	Selects the closest neighbour for each object
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		list		A list of objects
*		neig		A list of segments (distance objects)
*
*	Return value:
*		The number of segments
*/

int		Neighbour( DLINK *list, DLINK *neig );

/*
*	Selects the closest neighbour for each object
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		list		A list of objects
*		rats		A list of segments (distance objects)
*
*	Return value:
*		The number of segments
*/

int		Ratsnest( DLINK *list, DLINK *rats );

#endif
