/****************************************************************************/
/* Copyright 2004, © Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	GUI option box creation													*/
/*	~~~~~~~~~~~~~~~~~~~~~~~													*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zolt�n K�csi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#ifndef __OPTION_TYPES_H__
#define	__OPTION_TYPES_H__

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>

#include <types.h>
#include <options.h>

/****************************************************************************/
/****************************************************************************/

#define	OPT_STRLEN		10

/*
*	Types of options. Some are only decorative. The ones that are not,
*	are Flag, indicating a TRUE/FALSE choice, the List, that is a choice
*	from a list of items; Slid, which is a value from a continuous
*	range (graphically represented by a slider) and Spin, a spinbox.
*/

typedef	enum {

	otiHbox,						/* Option is a horizontal box			*/
	otiVbox,						/* Option is a vertical box				*/
	otiEbox,						/* End of packing box					*/
	otiSepl,						/* GUI separator line					*/
	otiHead,						/* GUI decorative header				*/
	otiLabl,						/* GUI decorative label					*/
	otiFlag,						/* Options is a toggle (on/off) thing	*/
	otiSlid,						/* Option controlled by a slider		*/
	otiSpin,						/* Option is a spinbox					*/
	otiList,						/* Option is a list	menu				*/
	otiCsel,						/* Option is a colour selector			*/
	otiCset,						/* Option is a colour settings			*/
	otiLast
	
} typOti;

/*
*	An option menu item is described by a structure. This structure can hold
*	not only a real option, but also items that are part of the GUI. In this
*	way a single array can describe both the options and their graphical
*	representation.
*	The type of the option tells what sort of thing it is (or if it is just
*	a GUI element, not a real option).
*	If it is a real option, then its name for the GUI and for the command 
*	line are specified. Also, the index into the options' value array should 
*	be given. In addition, if the option has some (type dependent) information
*	associated with it, a pointer points to that.
*/

typedef	struct s_opt {

	typOti	type;						/* Type of the item					*/
	char	*gui;						/* Item's name on the GUI			*/
	char	*cli;						/* Item's name for the CLI			*/
	typOpt	optid;						/* The option itself				*/
	void	*desc;						/* Type dependent descriptor		*/
	
} OPTION;

/*
*	Since sliders can be all sorts, logarithmic, quantised, interlocked
*	and so on. Therefore, the option itself has to supply functions that
*	deal with these things.
*	A slider has 4 functions, that form two pairs of forward-inverse
*	functions.
*	One pair transforms the slider value reported by GTK (a double,
*	in [0,1]) to a string (and back). That string will be displayed
*	beside the slider and stored as the current value of the slider.
*	The second pair of functions transforms the string to a value
*	that the option's native representation is (and back). These 
*	functions are called whenever the user wants to retrieve the option's
*	value or want to set it to a specified value.
*	A fifth, optional function can be declared that will be called
*	whenever the slider value changes. The function will be called 
*	when the value is already updated.
*/

typedef	void	(*SLIDPOSSTR)( OPTION *s, char *str, double pos );
typedef	double	(*SLIDSTRPOS)( OPTION *s, char *str );

typedef	void	(*SLIDSTRVAL)( OPTION *s, char *str, void *val );
typedef	void	(*SLIDVALSTR)( OPTION *s, char *str, void *val );
typedef	void	(*SLIDACTION)( OPTION *s );

/*
*	A slider is described by a structure that contains the four conversion
*	functions plus private data that is related to GUI functionality.
*/

typedef	struct {

	// Things that should be filled by the user
	
	SLIDSTRVAL		str_val;			/* Function: string   -> value		*/
	SLIDVALSTR		val_str;			/* Function: value    -> string		*/
	SLIDSTRPOS		str_pos;			/* Function: string	  -> position	*/
	SLIDPOSSTR		pos_str;			/* Function: position -> string		*/
	SLIDACTION		useract;			/* Optional user activity at change	*/
	
	// GTK related private fields
	
	GtkObject		*adjus;				/* Widget for the outermost frame	*/
	GtkWidget		*label;				/* Widget for the value label		*/
	GtkWidget		*slide;				/* Widget for the slide itself		*/

} SLIDER;

/*
*	A slider is described by a structure that contains the four conversion
*	functions plus private data that is related to GUI functionality.
*/

typedef	struct {

	int				min, max;			// Min and max values				*/
	void	(*func)( OPTION *, int );	// Change callback
		
	// GTK related private fields
	
	GtkObject		*adjus;				/* Widget for the outermost frame	*/
	GtkWidget		*spbox;				/* Widget for the slide itself		*/

} SPINNER;

/*
*	A menu item's value for the user is always an int (or enum).
*	The GUI and CLI know the item by a string. Each menu item 
*	consists a pointer to this triplet and a GUI related field.
*	The extra indirection if for cases when the the same menu 
*	is used multiple times.
*/

typedef	struct {

	// Things that should be filled by the user
	
	char		*gui;					/* GUI name for the option			*/
	char		*cli;					/* CLI name of the option			*/
	int			val;					/* Value for the option				*/

} CHOICE;

typedef struct {

	// Things that should be filled by the user

	CHOICE		*cid;					/* Choice descriptor				*/
		
	// GTK related private fields
	
	GtkWidget	*wid;					/* Widget for the button			*/
	
} ITEM;

typedef	struct {

	// Things that should be filled by the user

	ITEM		*items;					/* The menu choices					*/
	
	// GTK private fields
	
	GtkWidget	*frame;					/* Container frame					*/
	GtkWidget	*menu;					/* Menu option itself				*/
	GtkWidget	*list;					/* List of options					*/
	GSList		*group;					/* Group for the options			*/
	
} MENU;

typedef	struct {

	GtkWidget	*dialog;				/* Dialog box						*/
	GtkWidget	*canvas;				/* Canvas for the current colour	*/
	GtkWidget	*button;				/* Change button					*/
	int			colour;					/* Current colour value				*/

} CSEL;

/*
*	Initialiser macros
*/

#define	DEF_ITEM( name ) { &name, NULL }
#define	DEF_LAST		 { NULL, NULL  }
#define DEF_MENU( name ) { name, NULL, NULL, NULL, NULL };
#define DEF_SLID( f1, f2, f3, f4, f5 ) { f1, f2, f3, f4, f5, NULL, NULL, NULL }
#define	DEF_SPIN( f1, f2, f3 ) { f1, f2, f3, NULL, NULL }

/****************************************************************************/
/*						Data shared within functions						*/
/****************************************************************************/

extern	MENU	colour;				/* Colour separation menu descriptor	*/
extern	SLIDER	filter;				/* Filter kernel diameter slider desc.	*/
extern	SLIDER	gammac;				/* Gamma correction slider desc.		*/

extern	MENU	act_extr;			/* Active object extraction method		*/
extern	SLIDER	act_seed;			/* Active object seed value				*/
extern	SLIDER	act_acpt;			/* Active object acceptance value		*/
extern	SLIDER	act_size;			/* Active object minimum size			*/
		
extern	MENU	ref_extr;			/* Reference object extraction method	*/
extern	SLIDER	ref_seed;			/* Reference object seed value			*/
extern	SLIDER	ref_acpt;			/* Reference object acceptance value	*/
extern	SLIDER	ref_size;			/* Reference object minimum size		*/

extern	SLIDER	distmax;			/* Maximum distance						*/
extern	SPINNER	micron;				/* Maximum runtime						*/
extern	SPINNER	pixels;				/* Maximum runtime						*/

extern	CSEL	col_act;			/* Base colour for active objects		*/
extern	CSEL	col_ref;			/* Base colour for reference objects	*/

extern	CSEL	dis_col;			/* Distance vector colour				*/
extern	CSEL	act_col;			/* Active object colour					*/
extern	CSEL	ref_col;			/* Reference object colour				*/
extern	CSEL	con_col;			/* Confinement area colour				*/

extern	SLIDER	statsnum;			/* Number of statistics samples			*/
extern	SLIDER	stathist;			/* Number of histogram columns			*/
extern	MENU	stat_dis;			/* Type of segment calculation			*/

/****************************************************************************/
/*						Function prototypes									*/
/****************************************************************************/

/*
*	Builds the GUI from the array pointed by p
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Returns the GTK pointer to the toplevel widget.
*/

GtkWidget	*OptionsBuildGui( OPTION *p );

/*
*	Finds an option descriptor by the option's ID
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

OPTION	*OptionFindId( typOpt option );

/*
*	Returns the address of the option's value string by its ID
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

char	*OptionValue( typOpt option );


#endif
