/****************************************************************************/
/* Copyright 2004, © Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Preprocessing options													*/
/*	~~~~~~~~~~~~~~~~~~~~~													*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltán Kócsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <types.h>
#include <options.h>
#include <fatal.h>
#include "option_types.h"

/****************************************************************************/
/*							Local definitions								*/
/****************************************************************************/

#define	FILTER_MIN	0.0
#define	FILTER_MAX	20.0
#define	FILTER_RES	0.5

#define	GAMMA_UPPER	3.0
#define	GAMMA_LOWER	(1.0/GAMMA_UPPER)
#define	GAMMA_RES	21

/****************************************************************************/
/*								Prototypes									*/
/****************************************************************************/

static	void	FilterSV( OPTION *opt, char *str, void *val );
static	void	FilterVS( OPTION *opt, char *str, void *val );
static	double	FilterSP( OPTION *opt, char *str );
static	void	FilterPS( OPTION *opt, char *str, double val );
static	void	GammacSV( OPTION *opt, char *str, void *val );
static	void	GammacVS( OPTION *opt, char *str, void *val );
static	double	GammacSP( OPTION *opt, char *str );
static	void	GammacPS( OPTION *opt, char *str, double val );

/****************************************************************************/
/*					Data describing the option items						*/
/****************************************************************************/

static	CHOICE	crg	   = { "Light",			"lm",	OPT_COLOUR_SUBTRACT		};
static	CHOICE	cpb	   = { "Fluorescent",	"fm",	OPT_COLOUR_ADDITIVE		};
static	ITEM	col[]  = { DEF_ITEM( crg ), DEF_ITEM( cpb ), DEF_LAST };

	MENU	colour = DEF_MENU( col );
	SLIDER	filter = DEF_SLID( FilterSV, FilterVS, FilterSP, FilterPS, NULL );
	SLIDER	gammac = DEF_SLID( GammacSV, GammacVS, GammacSP, GammacPS, NULL );

/****************************************************************************/
/*					Filter kernel diameter slider functions					*/
/****************************************************************************/

/*
*	Converts a string to a user value
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FilterSV( OPTION *opt, char *str, void *val )
{
	if ( ! strcmp( str, "none" ) )
	
		*((double *) val) = 0;
	else
		sscanf( str, "%lf", val );
}

/*
*	Converts a user value to a string
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FilterVS( OPTION *opt, char *str, void *val )
{
double	d;
	
	d = *((double *) val);
	
	if ( d < FILTER_MIN ) d = FILTER_MIN;
	if ( d > FILTER_MAX ) d = FILTER_MAX;
	
	if ( ! d )
	
		strcpy( str, "none" );
	else
		sprintf( str, "% 4.1f", d );
}

/*
*	Converts a string into a position value
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	double	FilterSP( OPTION *opt, char *str )
{
double	val;

	FilterSV( opt, str, &val );
	if ( val < FILTER_MIN ) val = FILTER_MIN;
	if ( val > FILTER_MAX ) val = FILTER_MAX;
	val = (int) ( val / FILTER_RES ) * FILTER_RES;
	return( ( val - FILTER_MIN ) / ( FILTER_MAX - FILTER_MIN ) );
}

/*
*	Converts a position value to a string
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FilterPS( OPTION *opt, char *str, double val )
{
	val *= ( FILTER_MAX - FILTER_MIN );
	val  = (int)( val / FILTER_RES + 0.5 ) * FILTER_RES;
	
	if ( ! val )
	
		strcpy( str, "none" );
	else
		sprintf( str, "% 4.1f", val );
}

/****************************************************************************/
/*					Gamma correction slider functions						*/
/****************************************************************************/

/*
*	Converts a string to a user value
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	GammacSV( OPTION *opt, char *str, void *val )
{
	sscanf( str, "%lf", val );
}

/*
*	Converts a user value to a string
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	GammacVS( OPTION *opt, char *str, void *val )
{
double	d;
	
	d = *((double *) val);
	if ( d < GAMMA_LOWER ) d = GAMMA_LOWER;
	if ( d > GAMMA_UPPER ) d = GAMMA_UPPER;
	sprintf( str, "%4.2f", d );
}

/*
*	Converts a string into a position value
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	double	GammacSP( OPTION *opt, char *str )
{
double	d;

	GammacSV( opt, str, &d );
	if ( d < GAMMA_LOWER ) d = GAMMA_LOWER;
	if ( d > GAMMA_UPPER ) d = GAMMA_UPPER;
	
	d = log( d ) / log( GAMMA_UPPER ) / 2 + 0.5;
	return( d );
}

/*
*	Converts a position value to a string
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	GammacPS( OPTION *opt, char *str, double val )
{
	val = 2 * ( val - 0.5 );
	val = (int)( val * GAMMA_RES ) / (double) GAMMA_RES;
	val = exp( val * log( GAMMA_UPPER ) );
	sprintf( str, "%4.2f", val );
}
