/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Colour separates an image												*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~												*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <math.h>

#include <types.h>
#include <image.h>
#include <process.h>
#include <options.h>
#include <progress.h>

/****************************************************************************/
/*						Local types and constants							*/
/****************************************************************************/

#define	CSEL_PB_THOLD1	50
#define	CSEL_PB_THOLD2	20

#define	min( a, b )		((a)<(b)?(a):(b))
#define	max( a, b )		((a)>(b)?(a):(b))
#define	abs( a )		((a)<0?-(a):(a))

typedef	union {
	
	struct {
		
		double r, g, b;
	
	} rgb;
	
	struct {
	
		double h, s, v;
	
	} hsv;

} PIX;

/****************************************************************************/
/*								Prototypes									*/
/****************************************************************************/

static	PIX		RgbToHsv( PIX rgb );
static	IMAGE	*CselSubstr( IMAGE *old, int ric, int aic );
static	IMAGE	*CselAddit( IMAGE *old, int ric, int aic );

/****************************************************************************/
/*								Inlines										*/
/****************************************************************************/

/*
*	Checks a colour against a reference hue within a radius
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Returns the membership value. A pure grey (hue=0) is a
*	full member of any colour.
*	Hue is in the range of (0,6] for any colour that has a
*	non-zero saturation and 0 for a shade of grey.
*	Membership is 1 within half radius, linearly diminishes from
*	half to full radius and zero outside.
*/

static	inline	double	MapColour( double hue, double ref, double rad )
{
	if ( ! hue ) return( 1.0 );
	hue = abs( hue - ref );
	if ( hue > 3 ) hue = 6 - hue;
	if ( hue < rad / 2 ) return( 1.0 );
	if ( hue > rad ) return( 0.0 );
	return( 2 * ( rad - hue ) / rad );
}

/****************************************************************************/
/*							Public functions								*/
/****************************************************************************/

/*
*	Picks the colours according to the selector value
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*
*		old:	Original image
*		method:	Colour separation method
*		ric:	Reference feature colour in an int (3 bytes:00rrggbb)
*		aic:	Active feature colour in an int (3 bytes:00rrggbb)
*	
*	Return value
*
*		A pointer to the new image or NULL if could not finish
*
*	Receives an original image and returns a new one, the same size,
*	but without the B layer. The resulting image will contain the
*	active objects on the R layer and the reference on the G.
*	If the opration was aborted, returns NULL.
*/

IMAGE	*ColourSepImage( IMAGE *old, int method, int ric, int aic )
{
IMAGE	*new;

	switch ( method ) {
	
		case OPT_COLOUR_SUBTRACT:	return( CselSubstr( old, ric, aic ) );
		case OPT_COLOUR_ADDITIVE:	return( CselAddit( old, ric, aic ) );
		default:					return( NULL );
	}
}

/****************************************************************************/
/*							Local functions									*/
/****************************************************************************/

/*
*	RGB->HSV transformation
*	~~~~~~~~~~~~~~~~~~~~~~~
*	It is the method published by Smith and Lyons in the 
*	ACM Communications in 1996.
*	Hue is returned in (0,6], but it returns 0 if saturation is 0.
*	Pure red always generates a hue of 6.
*	R, G, B, S and V are all in [0,1].
*/

static	PIX		RgbToHsv( PIX rgb )
{
PIX		hsv;
double	w, v;
double	f, i;

	w = MIN( rgb.rgb.r, MIN( rgb.rgb.g, rgb.rgb.b ) );
	v = MAX( rgb.rgb.r, MAX( rgb.rgb.g, rgb.rgb.b ) );
	
	if ( v == w ) {
	
		// Zero saturation (grey), no H can be calculated
		// In that case we return 0
		
		hsv.hsv.h = 0;
		hsv.hsv.s = 0;
		hsv.hsv.v = v;
	}
	else {
	
		if ( w == rgb.rgb.r ) {
		
			// Brightest colour is red
			
			f = rgb.rgb.g - rgb.rgb.b;
			i = 3.0;
		}
		else if ( w == rgb.rgb.g ) {
		
			// Brightest colour is green
			
			f = rgb.rgb.b - rgb.rgb.r;
			i = 5.0;
		}
		else {
		
			// Brightest colour is blue
			
			f = rgb.rgb.r - rgb.rgb.g;
			i = 1.0;
		}
		
		hsv.hsv.h = i - f / ( v - w );
		hsv.hsv.s = ( v - w ) / v;
		hsv.hsv.v = v;
	}
	
	return( hsv );
}

/*
*	Colour separation for subtractive images
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	IMAGE	*CselSubstr( IMAGE *old, int ric, int aic )
{
IMAGE	*new;
jmp_buf	state;
PIX		pix;
PIX		refc, actc;
double	av, rv;
double	rad;
int		y, x;
int		xmax, ymax;
	
	new = ImageCreate( ymax = old->y, xmax = old->x, 2 );
	
	if ( setjmp( state ) ) {
	
		// Operation was cancelled 
		
		ImageDestroy( new );
		return( NULL );
	}
	
	actc.rgb.r = ( 255 - ( ( aic >> 16 ) & 255 ) ) / 255.0;
	actc.rgb.g = ( 255 - ( ( aic >>  8 ) & 255 ) ) / 255.0;
	actc.rgb.b = ( 255 - ( ( aic >>  0 ) & 255 ) ) / 255.0;
	
	refc.rgb.r = ( 255 - ( ( ric >> 16 ) & 255 ) ) / 255.0;
	refc.rgb.g = ( 255 - ( ( ric >>  8 ) & 255 ) ) / 255.0;
	refc.rgb.b = ( 255 - ( ( ric >>  0 ) & 255 ) ) / 255.0;
	
	actc = RgbToHsv( actc );
	refc = RgbToHsv( refc );
	
	rad = abs( actc.hsv.h - refc.hsv.h );
	if ( rad > 3 ) rad = 6 - rad;
		
	ProgressInit( "Colour transformation", state );
	
	for ( y = 0 ; y < ymax ; y++ ) {
	
		ProgressUpdate( y, old->y );
		
		for ( x = 0 ; x < xmax ; x++ ) {
		
			pix.rgb.r = ( 255 - old->im[ R ][ y ][ x ] ) / 255.0;
			pix.rgb.g = ( 255 - old->im[ G ][ y ][ x ] ) / 255.0;
			pix.rgb.b = ( 255 - old->im[ B ][ y ][ x ] ) / 255.0;
			
			pix = RgbToHsv( pix );
			
			av = MapColour( pix.hsv.h, actc.hsv.h, rad );
			rv = MapColour( pix.hsv.h, refc.hsv.h, rad );
						
			new->im[ R ][ y ][ x ] = av * pix.hsv.v * 255 + 0.5;
			new->im[ G ][ y ][ x ] = rv * pix.hsv.v * 255 + 0.5;
		}
	}
	
	ProgressDone();
	return( new );
}

/*
*	Colour separation for additive images
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	IMAGE	*CselAddit( IMAGE *old, int ric, int aic )
{
IMAGE	*new;
jmp_buf	state;
PIX		pix;
PIX		refc, actc;
double	av, rv;
double	rad;
int		y, x;
int		xmax, ymax;
	
	new = ImageCreate( ymax = old->y, xmax = old->x, 2 );
	
	if ( setjmp( state ) ) {
	
		// Operation was cancelled 
		
		ImageDestroy( new );
		return( NULL );
	}
	
	actc.rgb.r = ( ( aic >> 16 ) & 255 ) / 255.0;
	actc.rgb.g = ( ( aic >>  8 ) & 255 ) / 255.0;
	actc.rgb.b = ( ( aic >>  0 ) & 255 ) / 255.0;
	
	refc.rgb.r = ( ( ric >> 16 ) & 255 ) / 255.0;
	refc.rgb.g = ( ( ric >>  8 ) & 255 ) / 255.0;
	refc.rgb.b = ( ( ric >>  0 ) & 255 ) / 255.0;
	
	actc = RgbToHsv( actc );
	refc = RgbToHsv( refc );
	
	rad = abs( actc.hsv.h - refc.hsv.h );
	if ( rad > 3 ) rad = 6 - rad;

	ProgressInit( "Colour transformation", state );
	
	for ( y = 0 ; y < ymax ; y++ ) {
	
		ProgressUpdate( y, old->y );
		
		for ( x = 0 ; x < xmax ; x++ ) {
		
			pix.rgb.r = old->im[ R ][ y ][ x ] / 255.0;
			pix.rgb.g = old->im[ G ][ y ][ x ] / 255.0;
			pix.rgb.b = old->im[ B ][ y ][ x ] / 255.0;
			
			pix = RgbToHsv( pix );
			
			av = MapColour( pix.hsv.h, actc.hsv.h, rad );
			rv = MapColour( pix.hsv.h, refc.hsv.h, rad );
						
			new->im[ R ][ y ][ x ] = av * pix.hsv.v * 255 + 0.5;
			new->im[ G ][ y ][ x ] = rv * pix.hsv.v * 255 + 0.5;
		}
	}
	
	ProgressDone();
	return( new );
}
