/****************************************************************************/
/* Copyright 2004, © Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Object extraction - Delaunay triangulation								*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~								*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltán Kócsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <setjmp.h>

#include <types.h>
#include <memory.h>
#include <dlinklist.h>
#include <objects.h>
#include <process.h>
#include <progress.h>
#include "process_intern.h"

/****************************************************************************/
/*						Local definitions									*/
/****************************************************************************/

#define	SQR( x )	((x)*(x))
#define	ABS( x )	((x)<0?-(x):(x))

typedef	struct	{

	DLINK	link;
	OBJECT	ref;
	
} REFO;

/****************************************************************************/
/*						Public functions									*/
/****************************************************************************/

/*
*	Selects the lines that form the Delaunay mesh
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		list		A list of objects
*		tria		A list of segments (distance objects)
*
*	Return value:
*		The number of segments
*
*	The method is the following:
*	For each point pair, the line connecting the points belongs to
*	the Delaunay triangulation mesh iff there is a circle passing
*	through both points but its surface and circumference is free
*	of any other data point. To make it practical, we do this:
*	The smallest possible circle is the one of which the line
*	connecting Pi and Pj is a diameter. If there are no other 
*	points in or on that circle, then we have a solution and the
*	Pi-Pj line is part of the mesh. If there are points, then we 
*	have to test further. To see is we can construct a circle of
*	which Pi-Pj is a chord and it contains no other points, for 
*	each point k within the original circle we create a circle 
*	that passes throgh Pi, Pj and Pk. If there no forther points
*	in or on that circle, then by making it infinitesimally smaller
*	we can have a circle that also exludes Pk and this Pi-Pj is a
*	Delaunay line.
*	To save time we can do some speed-up. The Pi-Pj line cuts our
*	original circle into two semicircles. If there are points in
*	both semicircles then we can be sure that we will not be able
*	to construct a circle not containing any additional points, 
*	simply because if we make a circle that excludes points of one
*	semicircle of the original circle, it will always contain the
*	entire other half of the original circle. A further speedup is
*	that only one of the possible Pk circles have to be tested, the 
*	one of which the center is the farthest from the centre of the 
*	original circle.
*/

int		Delaunay( DLINK *list, DLINK *lines )
{
OBJECT	pi, pj, pk, ps;					// Points
DIST	*lp;							// Temp. line segment pointer
REFO	*tp;							// Temp. line segment pointer
DLINK	temp;							// Temp. line segment list
int		sp, sn;							// Counters for semicircles
double	ca, sa;							// Rotational angle sin and cos
double	xm, ym, rm, tmp;				// Minimal circle coords and radius
double	xl, yl;							// Large circle centre
double	a11, a12, a21, a22;				// Circle calc. coefficient matrix
double	c1, c2;							// Circle calc. constants
double	dt;								// Determinant
double	dk;								// Distance between circle centres
double	xs, ys, rs, ds;					// Selected circle params
double	xi, yi, xj, yj, xk, yk;			// Point coordinates
int		lc;								// Line segment count

	DlinkInit( lines );
	
	for ( lc = 0, pi = DlinkFirst( list ) ; pi ; pi = DlinkNext( pi ) ) {
	
		xi = pi->xc + pi->bbox.tlx;
		yi = pi->yc + pi->bbox.tly;
		
		for ( pj = DlinkNext( pi ) ; pj ; pj = DlinkNext( pj ) ) {
		
			xj = pj->xc + pj->bbox.tlx;
			yj = pj->yc + pj->bbox.tly;
			
			// Initialise the counters and other things
			
			DlinkInit( &temp );
			
			sp = 0;
			sn = 0;
						
			// Calculate the centre and radius of the minimal test circle
			
			xm = ( xi + xj ) / 2.0;
			ym = ( yi + yj ) / 2.0;
			rm = SQR( xm - xi ) + SQR( ym - yi );
			
			// We will have to decide whether points fall into one
			// semicircle or the other. The way we test them is that
			// if the Pi-Pj line is at angle alpha, then we rotate
			// the point being tested over Pi with -alpha and see if
			// its y coordinate is higher than that of Pi.
			// To rotate something with alpha, you have to multiply
			// it with the rotational matrix:
			//
			// | +cos(A) -sin(A) |   |x|   |x'|
			// |				 | x | | = |  |
			// | +sin(A) +cos(A) |   |y|   |y'|
			//
			// Calculate the sin and cos of the Pi-Pj line:
			
			sa = ( yj - yi ) / 2 / sqrt( rm );
			ca = ( xj - xi ) / 2 / sqrt( rm );
			
			// Examine all other points
			
			for ( pk = DlinkFirst( list ) ; pk ; pk = DlinkNext( pk ) ) {
			
				// Pi and Pj are simply skipped
				
				if ( pk == pi || pk == pj ) continue;
				
				xk = pk->xc + pk->bbox.tlx;
				yk = pk->yc + pk->bbox.tly;
				
				// Normally we should check in X and Y first and then
				// calculate the square but actually always calculating 
				// is quicker because we don't flush the insn pipe.
				
				if ( SQR( xk - xm ) + SQR( yk - ym ) > rm ) {
				
					// This points is outside of the circle
					
					continue;
				}
				
				// Pk is in the circle. Let's see which semicircle
				// it falls into. Since we only have to examine the
				// Y coordinate, we don't actually have to calculate
				// the entire matrix product. Rotate with -alpha:
				
				tmp = ( yk - yi ) * ca - ( xk - xi ) * sa;
				
				if ( ! tmp ) {
				
					// The point is on the line, or so nearly close that
					// the error is less than the pixelisation noise, so
					// we consider it to be on both sides.
					
					sp = 1;
					sn = 1;
				}
				else {				
				
					sp += tmp > 0;
					sn += tmp < 0;
				}
				
				// If we have at least one point on both sides, then we
				// have to work no further
				
				if ( sp && sn ) break;
				
				// There is still hope. We have to remember that this point
				// is within the semicircle. We use a REFO object to store it.
				
				tp = Malloc( sizeof( REFO ) );
				tp->ref = pk;
				DlinkAddTail( &temp, tp );								
			}
			
			// All points tested.
			
			if ( sp == 0 && sn == 0 ) {
			
				// No points within the circle, so Pi Pj is a Delaunay segment
				
				lp = MallocDistance();
				lp->pair.tlx = pi->xc + pi->bbox.tlx;
				lp->pair.tly = pi->yc + pi->bbox.tly;
				lp->pair.brx = pj->xc + pj->bbox.tlx;
				lp->pair.bry = pj->yc + pj->bbox.tly;
				lp->dist	 = SQR( lp->pair.tlx - lp->pair.brx ) +
							   SQR( lp->pair.tly - lp->pair.bry );
				DlinkAddTail( lines, lp );
				lc++;
			}
			else if ( sp == 0 || sn == 0 ) {
			
				// Points only on one side. We have to test them in turn.
				// The temp queue holds them.
				
				ds = 0;
				xs = 0;
				ys = 0;
				ps = NULL;
				
				while ( ( tp = DlinkRemoveFirst( &temp ) ) ) {
				
					pk = tp->ref;
					Free( tp );
					
					xk = pk->xc + pk->bbox.tlx;
					yk = pk->yc + pk->bbox.tly;
					
					// We have to calculate the circle that passes through
					// Pi, Pj, Pk. We have to solve the following equation:
					// 2 X (Xk-Xi) + 2 Y (Yk-Yi) = Yk^2 - Yi^2 + Xk^2 - Xi^2
					// 2 X (Xj-Xi) + 2 Y (Yj-Yi) = Yj^2 - Yi^2 + Xj^2 - Xi^2
					// We know that there is a solution (the 3 points are
					// not on one line) because then Pi Pj would not be
					// considered.
					
					a11 = 2 * ( xk - xi );
					a12 = 2 * ( yk - yi );
					a21 = 2 * ( xj - xi );
					a22 = 2 * ( yj - yi );
					c1  = SQR( yk ) - SQR( yi ) + SQR( xk ) - SQR( xi );
					c2  = SQR( yj ) - SQR( yi ) + SQR( xj ) - SQR( xi );
					
					// Get the matrix determinant
					
					dt = a11 * a22 - a12 * a21;
					
					// The 2x2 inverse matrix is:
					// 
					// | a11  a12 |-1    1    | a22  -a12 |
					// |          |   = --- * |           |
					// | a21  a22 |     det   | -a21  a11 |
					//
					// So, the solution (the circle centre) is:
					
					xl = ( a22 * c1 - a12 * c2 ) / dt;
					yl = ( a11 * c2 - a21 * c1 ) / dt;
					
					// Let's see if this circle is larger than any
					// before them
					
					dk = SQR( xl - xm ) + SQR( yl - ym );
					
					if ( dk > ds ) {
					
						// Remember this point
						
						xs = xl;
						ys = yl;
						ps = pk;
						ds = dk;
					}
				}
					
				// xs,ys are the centre of the big circle and ps is the
				// point that we selected as Pk.
				
				xk = ps->xc + ps->bbox.tlx;
				yk = ps->yc + ps->bbox.tly;
				
				// Calculate the square of the radius of the selected circle
					
				rs = SQR( xs - xk ) + SQR( ys - yk );
				
				// Now run through all points and see if they are within
				
				for ( pk = DlinkFirst( list ) ; pk ; pk = DlinkNext( pk ) ) {
				
					if ( pk == pi || pk == pj || pk == ps ) {
					
						// These points hold the circle, ignore them
						
						continue;
					}
				
					xk = pk->xc + pk->bbox.tlx;
					yk = pk->yc + pk->bbox.tly;
					
					if ( SQR( xk - xs ) + SQR( yk - ys ) <= rs ) {
					
						// This point is within, sp Pi - Pj can not be a 
						// Delaunay segment.
						
						break;
					}
				}
				
				if ( ! pk ) {
				
					// Pi - Pk is a valid segment, add it to the list
					
					lp = MallocDistance();
					lp->pair.tlx = pi->xc + pi->bbox.tlx;
					lp->pair.tly = pi->yc + pi->bbox.tly;
					lp->pair.brx = pj->xc + pj->bbox.tlx;
					lp->pair.bry = pj->yc + pj->bbox.tly;
					lp->dist	 = SQR( lp->pair.tlx - lp->pair.brx ) +
								   SQR( lp->pair.tly - lp->pair.bry );
					DlinkAddTail( lines, lp );
					lc++;
				}
			}
			else {
			
				// Nothing to do, just free whatever we have stored
				
				while ( ( tp = DlinkRemoveFirst( &temp ) ) ) Free( tp );
			}
		}
	}
	
	return( lc );
}
