/****************************************************************************/
/* Copyright 2004, © Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Object extraction - Dijkstra's minimal spanning tree algorithm			*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~			*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltán Kócsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <setjmp.h>

#include <types.h>
#include <memory.h>
#include <dlinklist.h>
#include <objects.h>
#include <process.h>
#include <progress.h>
#include "process_intern.h"

/****************************************************************************/
/*						Local definitions									*/
/****************************************************************************/

#define	SQR( x )	((x)*(x))
#define	ABS( x )	((x)<0?-(x):(x))

typedef	struct	{

	int		this_x, this_y;
	int		from_x, from_y;
	int		dist;
	
} VERTEX;

/****************************************************************************/
/*						Public functions									*/
/****************************************************************************/

/*
*	Selects the lines that form the minimum cost spanning tree
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		list		A list of objects
*		tria		A list of segments (distance objects)
*
*	Return value:
*		The number of segments
*
*	The method is Dijkstra's algorithm on a fully connected graph:
*	Select a vertex, this will be the starting segment.
*	This vertex is to be removed from the working set.
*	Calculate the distances of all vertices from the starting one.
*	Note for each vertex that the distance currently hold is the length
*	of the edge between the vertex and the origin vertex.
*	Then, until all vertices are removed, do the following:
*	Select the vertex with the sortest distance. The edge between the
*	selected vertex and the vertex noted as its reference vertex is
*	added to the result. For all vertices, calculate the distance 
*	between the vertex and the selected vertex. If the distance is smaller
*	than the noted distance between the vertex and the original, then
*	make the selected vertex be the reference for the vertex and update
*	the distance for the vertex to the shorter value. Remove the selected
*	vertex from the set of vertices.
*/

int		Dijkstra( DLINK *list, DLINK *tria )
{
int		numv;
VERTEX	*vert, temp;
OBJECT	op;
int		d, i, j, k;
DIST	*segm;

	DlinkInit( tria );
	
	// Calculate the number of vertices and allocate the vertex array
	
	for ( numv=0, op=DlinkFirst( list ) ; op ; op=DlinkNext( op ), numv++ );
	vert = Malloc( sizeof( VERTEX ) * numv );
	
	// Initialise the vertices
	
	for ( i = 0, op = DlinkFirst( list ) ; op ; op=DlinkNext( op ), i++ ) {
	
		vert[ i ].this_x = op->xc + op->bbox.tlx;
		vert[ i ].this_y = op->yc + op->bbox.tly;
		vert[ i ].dist   = DIST_INF;
	}
	
	// Calculate the distance of the vertices from the start vertex
		
	for ( i = 1 ; i < numv ; i++ ) {
	
		vert[ i ].from_x = vert[ 0 ].this_x;
		vert[ i ].from_y = vert[ 0 ].this_y;
		vert[ i ].dist	 = SQR( vert[ 0 ].this_x - vert[ i ].this_x ) +
						   SQR( vert[ 0 ].this_y - vert[ i ].this_y );
	}
	
	// Start Dijkstra's algorithm
	
	for ( i = 1 ; i < numv ; i++ ) {
	
		// Find the vertex closest to the source
		
		d = DIST_INF;
		k = 0;
			
		for ( j = i ; j < numv ; j++ ) {
		
			if ( vert[ j ].dist < d ) {
			
				d = vert[ j ].dist;
				k = j;
			}
		}

		// Create a line segment and add it to the result
		
		segm = MallocDistance();
		segm->pair.tlx = vert[ k ].this_x;
		segm->pair.tly = vert[ k ].this_y;
		segm->pair.brx = vert[ k ].from_x;
		segm->pair.bry = vert[ k ].from_y;
		segm->dist	   = vert[ k ].dist;
		DlinkAddTail( tria, segm );
		
		// Move the selected vertex to the done set
		
		temp = vert[ i ];
		vert[ i ] = vert[ k ];
		vert[ k ] = temp;
		
		// Update the distances of shortests paths
		
		for ( j = i + 1 ; j < numv ; j++ ) {
	
			d = SQR( vert[ i ].this_x - vert[ j ].this_x ) +
				SQR( vert[ i ].this_y - vert[ j ].this_y );
			
			if ( d < vert[ j ].dist ) {
				vert[ j ].from_x = vert[ i ].this_x;
				vert[ j ].from_y = vert[ i ].this_y;
				vert[ j ].dist	 = d;
			}
		}
	}
	
	Free( vert );
	return( numv - 1 );
}
