/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Calculating distances													*/
/*	~~~~~~~~~~~~~~~~~~~~~													*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <types.h>
#include <dlinklist.h>
#include <objects.h>
#include <distance.h>
#include <memory.h>
#include "process_intern.h"
#include "distfunc.h"

/****************************************************************************/
/*						Local types and constants							*/
/****************************************************************************/

typedef	struct {

	OBJECT	ref;				/* Reference object							*/
	OBJECT	act;				/* Active object							*/
	int		val;				/* Distance between the objects				*/
	
	union	{
		
		RECT	pair;			/* Pixel pair								*/
		DLINK	list;			/* Pixel list								*/
	
	} res;
	
} DISTMAT;

typedef	struct {

	DLINK	*ref;				/* List of reference objects				*/
	DLINK	*act;				/* List of active objects					*/
	DISTMAT	**mat;				/* Distance matrix							*/
	int		trp;				/* Matrix is transposed						*/
	int		col;				/* Number of columns in matrix				*/
	int		row;				/* Number of rows in matrix					*/
	int		dismax;				/* Maximum allowed distance					*/
	int		actmax;				/* Maximum pairs for active					*/
	int		refmax;				/* Maximum pairs for reference				*/
	int		pixlst;				/* Needs the pixel list						*/
	
} MAPINFO;

typedef	struct {

	int		dis;				/* Distance									*/
	int		row;				/* Matrix row index							*/
	int		col;				/* Matrix column index						*/
	
} DISTREF;

#define	START_PROB			(RAND_MAX/8)
#define	COOLING_SPEED		100
#define	COOLING_SHIFT		1
#define	REHEAT_TRIES		(MAX_TRIES/20)
#define	MAX_TRIES			300000

/****************************************************************************/
/*						Prototypes											*/
/****************************************************************************/

static	void	DistMapMatrixCreate( MAPINFO *map );
static	void	DistMapMatrixValues( MAPINFO *map );
static	void	DistMapMatrixDelete( MAPINFO *map );
static	void	DistMapRef( MAPINFO *mat, DLINK *dlst, DLINK *olst );
static	void	DistMapAct( MAPINFO *mat, DLINK *dlst, DLINK *olst );
static	void	DistMapAll( MAPINFO *mat, DLINK *dlst, DLINK *olst );
static	void	DistMapOne( MAPINFO *mat, DLINK *dlst, DLINK *olst );
static	int		DistMapCmp( const void *a, const void *b );

/****************************************************************************/
/*						Public functions									*/
/****************************************************************************/

/*
*	Maps the shortest distance between an object and a list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters:
*		list		A list of objects to compare with
*		obj			A single object to compare against list members
*		mind		An acceptable minimum solution
*		maxd		An upper limit for the solution distance
*
*	Return value:
*		A pointer to a newly allocated DIST object or NULL
*
*	Stops mapping if the minimum criteria is satisfied.
*	Returns a DIST pointer if it found a pair less than
*	the maximum apart or NULL if not.
*/

DIST	*DistMapObjList( DLINK *list, OBJECT obj, int mind, int maxd )
{
OBJECT	p;
RECT	rect;
DIST	*ptr;
int		act;
int		min;

	ptr = MallocDistance();
	ptr->dist = DIST_INF;
	
	for ( min = DIST_INF, p = DlinkFirst( list ) ; p ; p = DlinkNext( p ) ) {
	
		act = DistMin( p, obj, &rect, maxd );
		
		if ( act < maxd ) {
		
			maxd	  = act;
			ptr->pair = rect;
			ptr->dist = act;
			
			if ( act <= mind ) break;
		}
	}
	
	if ( ptr->dist  != DIST_INF ) {
		
		return( ptr );
	}
	else {
	
		Free( ptr );
		return( NULL );
	}
}

/*
*	Maps the distance between two lists
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters:
*		dis		An empty list onto which the result DIST pairs will be put
*		ovl		An empty list onto which the result OVLP pairs will be put
*		ref		List of reference objects
*		act		List of active objects
*		maxdis	Maximum acceptable distance between objects
*		actmul	Flag indicating that active objects can have multiple pairs
*		refmul	Flag indicating that reference objects can have multiple pairs
*		pixlst	Flag indicating is the OVLP pixel list is needed or not
*
*	Return value:
*		TRUE is succeeded ABORT if not
*/

int		DistMapLists( DLINK *dis, DLINK *ovl, DLINK *ref, DLINK *act,
					  int maxdis, int refmul, int actmul, int pixlst )
{
MAPINFO	map;

	map.ref		= ref;
	map.act		= act;
	map.dismax	= maxdis * maxdis;
	map.actmax	= actmul;
	map.refmax	= refmul;
	map.pixlst	= pixlst;
	DlinkInit( dis );
	DlinkInit( ovl );
	
	DistMapMatrixCreate( &map );
	DistMapMatrixValues( &map );
	
	if ( refmul ) {
	
		if ( actmul ) {
		
			DistMapAll( &map, dis, ovl );
		}
		else {
		
			DistMapAct( &map, dis, ovl );
		}
	}
	else {
	
		if ( actmul ) {
		
			DistMapRef( &map, dis, ovl );
		}
		else {
		
			DistMapOne( &map, dis, ovl );
		}
	}
	
	DistMapMatrixDelete( &map );
	return( TRUE );
}

/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/

/*
*	Create the initial pairing matrix
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	DistMapMatrixCreate( MAPINFO *map )
{
int		ac, rc, i;
DLINK	*al, *rl;
OBJECT	ap, rp;
DISTMAT	**mat;
	
	// Count the number of elements
		
	for ( ac = 0, ap = DlinkFirst(map->act) ; ap ; ap = DlinkNext(ap), ac++ );
	for ( rc = 0, rp = DlinkFirst(map->ref) ; rp ; rp = DlinkNext(rp), rc++ );

	// If the number of actuals is greather than the number of
	// references, we swap them
	
	if ( ac > rc ) {
	
		i  = ac;
		ac = rc;
		rc = i;
		al = map->ref;
		rl = map->act;

		map->trp = TRUE;
	}
	else {
	
		al = map->act;
		rl = map->ref;
		
		map->trp = FALSE;
	}
	
	// Now we can be sure that rc >= ac. Create what we need:
	// Create an [rc][ac] matrix of distances.
	// Note, that we treet the matrix as mat[ columns ][ rows ]
	
	mat = MallocArray( rc, ac, sizeof( DISTMAT ) );
	
	// Do an initial fill of the matrix
	
	for ( rc = 0, rp = DlinkFirst(rl) ; rp ; rp = DlinkNext(rp), rc++ ) {
		
		for ( ac = 0, ap = DlinkFirst(al) ; ap ; ap = DlinkNext(ap), ac++ ) {
		
			mat[ rc ][ ac ].ref = rp;
			mat[ rc ][ ac ].act = ap;
		}
	}
	
	// Put things into the mapinfo
	
	map->mat = mat;
	map->col = rc;
	map->row = ac;
}

/*
*	Fills the matrix with distances and overlaps
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Overlaps are only calculated when needed. They are treated as negative
*	distances, which assures that overlaps are actually considered smaller
*	than any real distance.
*/

static	void	DistMapMatrixValues( MAPINFO *map )
{
int		col, row, dist;
DLINK	*lptr;

	for ( col = 0 ; col < map->col ; col++ ) {
	
		for ( row = 0 ; row < map->row ; row++ ) {
	
			dist = DistMin( map->mat[ col ][ row ].ref, 
							map->mat[ col ][ row ].act,
							&map->mat[ col ][ row ].res.pair, 
							map->dismax );
			
			if ( dist == 0 ) {
			
				DlinkInit( &map->mat[ col ][ row ].res.list );
				lptr = map->pixlst ? &map->mat[ col ][ row ].res.list : NULL;
				dist = -DistOvl( map->mat[ col ][ row ].ref, 
								 map->mat[ col ][ row ].act,
								 lptr );
			}
			
			map->mat[ col ][ row ].val = dist;
		}
	}
}

/*
*	Releases the matrix
*	~~~~~~~~~~~~~~~~~~~
*/

static	void	DistMapMatrixDelete( MAPINFO *map )
{
int		col, row, dist;
PIXEL	p;

	for ( col = 0 ; col < map->col ; col++ ) {
	
		for ( row = 0 ; row < map->row ; row++ ) {
		
			if ( map->mat[ col ][ row ].val < 0 ) {
			
				while (( p = DlinkRemoveFirst( &map->mat[col][row].res.list )))
				
					Free( p );
			}
		}
	}
	
	Free( map->mat );
}

/*
*	Selects the smallest pair for each reference object
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	DistMapRef( MAPINFO *map, DLINK *dlst, DLINK *olst )
{
DIST	*dist;
OVLP	*ovlp;
int		min, sel;
int		rc, ac;

	if ( map->trp ) {
	
		map->trp = FALSE;
		DistMapAct( map, dlst, olst );
		return;
	}
	
	for ( rc = 0 ; rc < map->col ; rc++ ) {
	
		min = DIST_INF;
		sel = -1;
				
		for ( ac = 0 ; ac < map->row ; ac++ ) {
		
			if ( map->mat[ rc ][ ac ].val < min ) {
			
				min = map->mat[ rc ][ ac ].val;
				sel = ac;
			}
		}
		
		if ( min != DIST_INF ) {
		
			if ( min < 0 ) {
				
				ovlp = MallocOverlap();
				DlinkInit( &ovlp->pixs );
				DlinkCat( &ovlp->pixs, &map->mat[ rc ][ sel ].res.list );
				ovlp->nump = -min;
				DlinkAddTail( olst, ovlp );
			}
			else {
			
				dist = MallocDistance();
				dist->dist = min;
				dist->pair = map->mat[ rc ][ sel ].res.pair;
				DlinkAddTail( dlst, dist );
			}
		}
	}
}

/*
*	Selects the smallest pair for each active object
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	DistMapAct( MAPINFO *map, DLINK *dlst, DLINK *olst )
{
DIST	*dist;
OVLP	*ovlp;
int		min, sel;
int		rc, ac;

	if ( map->trp ) {
	
		map->trp = FALSE;
		DistMapRef( map, dlst, olst );
		return;
	}
	
	for ( ac = 0 ; ac < map->row ; ac++ ) {
	
		min = DIST_INF;
		sel = -1;
		
		for ( rc = 0 ; rc < map->col ; rc++ ) {
		
			if ( map->mat[ rc ][ ac ].val < min ) {
			
				min = map->mat[ rc ][ ac ].val;
				sel = rc;
			}
		}
		
		if ( min != DIST_INF ) {
		
			if ( min < 0 ) {
				
				ovlp = MallocOverlap();
				DlinkInit( &ovlp->pixs );
				DlinkCat( &ovlp->pixs, &map->mat[ sel ][ ac ].res.list );
				ovlp->nump = -min;
				DlinkAddTail( olst, ovlp );
			}
			else {
			
				dist = MallocDistance();
				dist->dist = min;
				dist->pair = map->mat[ sel ][ ac ].res.pair;
				DlinkAddTail( dlst, dist );
			}
		}
	}
}

/*
*	Selects all valid pairs
*	~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	DistMapAll( MAPINFO *map, DLINK *dlst, DLINK *olst )
{
DIST	*dist;
OVLP	*ovlp;
int		rc, ac;

	for ( ac = 0 ; ac < map->row ; ac++ ) {
	
		for ( rc = 0 ; rc < map->col ; rc++ ) {
		
			if ( map->mat[ rc ][ ac ].val != DIST_INF ) {
		
				if ( map->mat[ rc ][ ac ].val < 0 ) {
				
					ovlp = MallocOverlap();
					DlinkInit( &ovlp->pixs );
					DlinkCat( &ovlp->pixs, &map->mat[ rc ][ ac ].res.list );
					ovlp->nump = -map->mat[ rc ][ ac ].val;
					DlinkAddTail( olst, ovlp );
				}
				else {
			
					dist = MallocDistance();
					dist->dist = map->mat[ rc ][ ac ].val;
					dist->pair = map->mat[ rc ][ ac ].res.pair;
					DlinkAddTail( dlst, dist );
				}
			}
		}
	}
}

/*
*	Selects pairs with a strictly one-to-one restriction
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	This is the hard case. If we create a matrix that contains the distances
*	between each pair from the two lists, the restrictions instruct us to
*	chose at most 1 element from each row and 1 from each column while
*	we have to reach minimum sum of chosen distances. This is an NP problem,
*	so we try to find a reasonably good solution in a reasonably short time.
*	
*	The algorithm consists of several steps. 
*	First, we already have the distance matrix in such a way that the 
*	number of columns is greater than or equal to the number of rows.
*	Now we create an array of those elements in the matrix that are 
*	non-infinite distances (i.e. valid).
*	We sort this array by distance. Then starting at the smallest item, we
*	choose pairs in such a way that each row and each column can be used
*	at most once. This is a greedy selection and is used to generate the
*	initial arrangement.
*	
*	We create an array. Each element belongs to a column and holds which
*	element within the column is currently selected (each column can only
*	have one element). If a column is empty, we mark it with a special value.
*
*	Then, in a loop, we select two columns in random.
*	If both of them are empty, we discard the selection and go again.
*	If both are in use, then we try to swap the elements (this will not
*	violate the one per rox/one per column restriction). If neither of the
*	new distances if infinite, we might have a chance. We check the sum 
*	distance after the change, and if it is smaller, we accept it.
*	If one column is empty, then instead of swapping elements, we just
*	try to move the element from the used column to the empty one (and
*	release the full one).
*	We might allow some deterioration of the distance, (aka simulated
*	annealing) by a constantly decreasing probability, it gets the 
*	algorithm out of places where it would usually stuck.
*	If the annealing seems to be stuck even this way, we 'reheat' it by 
*	increasing the acceptance probability slightly.
*/

static	void	DistMapOne( MAPINFO *map, DLINK *dlst, DLINK *olst )
{
int		n, i, j;
int		count, tries;
int		c1, r1, c2, r2;
int		maxd, gain, prob;
int		ac, rc;
DISTREF	*lst;
int		*col, *row;
DIST	*dist;
OVLP	*ovlp;

	// We know that col >= row. Create what we need:
	// - an col*row array of (rx, ax) index pairs for sorting
	// - two col vectors to hold counters for the greedy selection
	//	 and to hold the occupancy and best results for the annealing
	
	col = Malloc( map->col * sizeof( int ) );
	row = Malloc( map->col * sizeof( int ) );
	lst = Malloc( map->col * map->row * sizeof( DISTREF ) );
	
	// Now we fill the the list all valid distances
	
	n = 0;
	
	for ( rc = 0 ; rc < map->col ; rc++ ) {
	
		for ( ac = 0 ; ac < map->row ; ac++ ) {
		
			if ( map->mat[ rc ][ ac ].val != DIST_INF ) {
			
				lst[ n ].col = rc;
				lst[ n ].row = ac;
				lst[ n ].dis = map->mat[ rc ][ ac ].val;
				n++;
			}
		}
	}

	// This is just for typing convenience
	
	rc = map->col;
	ac = map->row;
	
	// Now we sort the distance list in order to create an initial
	// arrangement
	
	qsort( lst, n, sizeof( DISTREF ), DistMapCmp );
	
	// Find an initial solution
	
	memset( col, 0, sizeof( int ) * rc );
	memset( row, 0, sizeof( int ) * rc );
	
	for ( i = 0, j = 0 ; i < n && j < rc ; i++ ) {
	
		if ( ! col[ lst[ i ].col ] && ! row[ lst[ i ].row ] ) {
		
			col[ lst[ i ].col ] = lst[ i ].row + 1;
			row[ lst[ i ].row ] = lst[ i ].col + 1;
			j++;
		}
	}
	
	// At this moment we have an initial solution. col[ n ] contains
	// for each column the selected item's row number + 1 or 0, if
	// the given column is not used. Save this solution as our base one.
	// We use the row array as storage for the best solution found so far
	
	memcpy( row, col, rc * sizeof( int ) );
	
	for ( maxd = 0, i = 0 ; i < rc ; i++ ) {
	
		if ( col[ i ] ) maxd += map->mat[ i ][ col[ i ] - 1 ].val;
	}
	
	// We are ready for the annealing
	
	prob = START_PROB;
		
	for ( count = tries = 0 ; tries < MAX_TRIES ; tries++ ) {
		
		// Select two columns in random. Normally, the modulus method
		// would be a disaster, but we use random() instead of rand(),
		// which guarantees good distribution on the lower bits as well.
		
		c1 = random() % rc;
		c2 = random() % rc;

		// They must be different
		
		if ( c1 == c2 ) continue;
		
		// If both are empty, try again
		
		if ( ! col[ c1 ] && ! col[ c2 ] ) continue;
		
		// Calculate the corresponding rows. If both are taken, then
		// we will swap them, if only one is taken, we keep that row
		
		if ( col[ c1 ] && col[ c2 ] ) {
		
			r1 = col[ c1 ] - 1;
			r2 = col[ c2 ] - 1;
		}
		else {
		
			if ( ! col[ c1 ] )
		
				r1 = r2 = col[ c2 ] - 1;
			else
				r1 = r2 = col[ c1 ] - 1;
		}
		
		// Let's see if we can swap them at all
		
		if ( map->mat[ c1 ][ r2 ].val == DIST_INF || 
			 map->mat[ c2 ][ r1 ].val == DIST_INF ) {
		
			// No, we can't. Try again
			
			continue;
		}
		
		// Yes, a swap is possible, so we consider it as a try. We increment
		// the try count *again* because we are going to spend some time here.
		
		tries++;
		
		// Update annealing things.
		// If we didn't have any gain in the past N counts, we try to
		// increase the probability of swap despite of a loss.
		
		count++;
		
		if ( count == REHEAT_TRIES ) {
		
			if ( prob >> COOLING_SHIFT )
			
				prob += prob >> COOLING_SHIFT;
			else
				prob++;		
		}
		else {
		
			if ( prob && tries % COOLING_SPEED == 0 ) {
		
				if ( prob >> COOLING_SHIFT == 0 )
			
					prob--;
				else
					prob -= prob >> COOLING_SHIFT;
			}
		}
		
		// Calculate the gain for the proposed change
		
		if ( ! col[ c1 ] ) {
		
			// r1 == r2 and we move from [c2][r1] to [c1][r1]
			
			gain = map->mat[ c2 ][ r1 ].val - map->mat[ c1 ][ r1 ].val;
		} 
		else if ( ! col[ c2 ] ) {
		
			// r1 == r2 and we move from [c2][r1] to [c1][r1]
			
			gain = map->mat[ c1 ][ r1 ].val - map->mat[ c2 ][ r1 ].val;
		} 
		else {
		
			// We swap them, that is [c1][r1] goes to [c2][r1] and 
			// [c2][r2] goes to [c1][r2]
			
			gain = ( map->mat[ c1 ][ r1 ].val - map->mat[ c2 ][ r1 ].val ) +
				   ( map->mat[ c2 ][ r2 ].val - map->mat[ c1 ][ r2 ].val );
		}

		if ( gain >= 0 || random() < prob ) {
		
			// We either have a definite gain or the annealing said
			// that we should accept this change.
			// Update everything. We have to fetch r1 and r2 again!
			
			r1 = col[ c1 ];
			r2 = col[ c2 ];
			col[ c1 ] = r2;
			col[ c2 ] = r1;
			row[ c1 ] = col[ c1 ];
			row[ c2 ] = col[ c2 ];
			maxd -= gain;
			count = 0;
		}
	}
	
	// OK, we finished trying, return the solution
	
	for ( i = 0 ; i < rc ; i++ ) {
	
		if ( row[ i ] ) {
		
			if ( map->mat[ i ][ row[ i ] - 1 ].val < 0 ) {
		
				ovlp = MallocOverlap();
				DlinkInit( &ovlp->pixs );
				DlinkCat(&ovlp->pixs, &map->mat[ i ][ row[ i ] - 1 ].res.list);
				ovlp->nump = -map->mat[ i ][ row[ i ] - 1 ].val;
				DlinkAddTail( olst, ovlp );
			}
			else {
			
				dist = MallocDistance();
				dist->dist = map->mat[ i ][ row[ i ] - 1 ].val;
				dist->pair = map->mat[ i ][ row[ i ] - 1 ].res.pair;
				DlinkAddTail( dlst, dist );
			}
		}
	}
	
	// Release resources
	
	Free( lst );
	Free( row );
	Free( col );
}

/*
*	Comparison function for quicksort
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	int		DistMapCmp( const void *a, const void *b )
{
	if ( ((DISTREF *) a)->dis < ((DISTREF *) b)->dis ) return( -1 );
	if ( ((DISTREF *) a)->dis > ((DISTREF *) b)->dis ) return( +1 );
	return( 0 );
}
