/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Object extraction														*/
/*	~~~~~~~~~~~~~~~~~~														*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <types.h>
#include <dlinklist.h>
#include <objects.h>
#include <memory.h>
#include <image.h>
#include <progress.h>
#include <options.h>
#include <process.h>
#include "process_intern.h"

/****************************************************************************/
/*							Local definitions								*/
/****************************************************************************/

#define	SCANSIZE		8			/* Scanning for blob max.				*/

typedef	struct	{

	DLINK	*list;					/* Extracted object list				*/
	int		**mark;					/* Memory block used for mapping		*/
	int		ys, xs;					/* Size of the image					*/
	int		blobid;					/* Blob identifier (counter)			*/
	ubyte	**layer;				/* Layer we're processing				*/
	int		min_slev;				/* Minimum seed level					*/
	int		min_alev;				/* Minimum acceptance level				*/
	int		min_pixs;				/* Minimum object size					*/
	int		progress;				/* Progress bar needs updating			*/
	
} INFO;

/****************************************************************************/
/*							Prototypes										*/
/****************************************************************************/

static	void	BlobSearch( INFO *info );
static	void	BlobScan( INFO *info, DLINK *blst, int lev );
static	void	BlobBlob( INFO *info, DLINK *blst );
static	void	BlobPix( INFO *info, OBJECT blob );
static	void	BlobTrim( INFO *info, OBJECT blob );
static	void	BlobUnpix( INFO *info, OBJECT blob, PIXEL p );
static	int		BlobGrow( INFO *info, OBJECT blob );
static	int		BlobShrink( INFO *info, OBJECT blob ); 
static	int		BlobNeighbour( INFO *info, int y, int x, int id );

static	void	TwigSearch( INFO *info );
static	int		TwigGrow( INFO *info, DLINK *twig, int y, int x );

/****************************************************************************/
/*							Public functions								*/
/****************************************************************************/

/*
*	Finds the objects on an image layer, using the selected algorithm
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*
*		list		A DLINK onto which the feature objects will be placed
*		image		The image on which the search should be done
*		layer		The layer of the image
*		conf		A list of objects representing excluded areas
*		prog		Title for the porgress bar
*		method		Extraction method
*		seed		Minimum seed value for feature growing
*		acpt		Minimum value to consider a pixel
*		size		Minimum feature size (for blobsearch only)
*	
*	Return value
*
*		TRUE if succeeded, FALSE if it was aborted.
*
*	It does not do a list trimming, just finds anything that is
*	acceptable by the search algorithm. Places the objects onto
*	the DLINK passed to it. The DLINK will be empty if aborted.
*/

int		ObjectSearch( DLINK *list, IMAGE *image, int layer, DLINK *conf,
					  char *prog, int method, int seed, int acpt, int size )
{
INFO	info;
OBJECT	obj;
PIXEL	pix;
jmp_buf	state;

	info.ys			= image->y;
	info.xs			= image->x;
	info.blobid		= 1;
	info.list		= list;
	info.layer		= image->im[ layer ];
	info.mark		= MallocArray( image->y, image->x, sizeof( int ) );
	info.min_slev	= seed;
	info.min_alev	= acpt;
	info.min_pixs	= size;
	info.progress	= TRUE;
		
	DlinkInit( info.list );
	
	if ( setjmp( state ) ) {
	
		// Aborted
		
		Free( info.mark );
		while ( ( obj = DlinkRemoveFirst( info.list ) ) ) Free( obj );
		return( FALSE );
	}
	
	ProgressInit( prog, state );
	
	// Before starting to search for objects, we have to mark all
	// areas that are excluded from the search
	
	for ( obj = DlinkFirst( conf ) ; obj ; obj = DlinkNext( obj ) ) {
	
		for ( pix=DlinkFirst( &obj->pixels ) ; pix ; pix=DlinkNext( pix ) ) {
		
			info.mark[ obj->bbox.tly + pix->y ][ obj->bbox.tlx + pix->x ] = -1;
		}
	}		
	
	switch ( method ) {
	
		case OPT_EXT_BLOB:
			
			BlobSearch( &info );
			break;
		
		case OPT_EXT_TWIG:
		
			TwigSearch( &info );
			break;
			
		default:
			break;	
	}

	ProgressDone();
	Free( info.mark );
	return( TRUE );
}

/*
*	Turns a bitmap into an object list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*
*	Parameters
*
*		list		A DLINK onto which the objects will be placed
*		bitmap		An array of chars, representing the bitmapped image
*		ys, xs		Size of the array
*
*	Return value
*		-
*
*	Each pixel is represented as a byte on the bitmap. If the byte is
*	non-zero, the pixel will be part of an object. The objects represent
*	continuous marked areas of the bitmap.
*/

void	ObjectMapped( DLINK *list, ubyte **bitmap, int ys, int xs )
{
INFO	info;
OBJECT	obj;
jmp_buf	state;

	ObjectDestroyList( list );
	
	info.ys			= ys;
	info.xs			= xs;
	info.blobid		= 1;
	info.list		= list;
	info.layer		= bitmap;
	info.mark		= MallocArray( ys, xs, sizeof( int ) );
	info.min_slev	= 1;
	info.min_alev	= 1;
	info.min_pixs	= 1;
	info.progress	= FALSE;
	TwigSearch( &info );
}

/****************************************************************************/
/*							Search for blobs								*/
/****************************************************************************/

/*
*	Searches for blobs on the given image layer
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	BlobSearch( INFO *info )
{
int		level;
DLINK	levellist;

	DlinkInit( &levellist );
	
	for ( level = 255 ; level > info->min_slev ; level-- ) {
		
		if ( info->progress ) ProgressUpdate( 255-level, 255-info->min_slev );
		BlobScan( info, &levellist, level );
		BlobBlob( info, &levellist );
		DlinkCat( info->list, &levellist );
	}
}

/*
*	Searches for blobs starting at the given level
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	A pixel will be a seed for a blob if it is at the requested level
*	and if there's no other starting point within the scan region.
*	If a pixel is decided that it will be a start point for a blob, 
*	a new blob is created and added to the blob list.
*/

static	void	BlobScan( INFO *info, DLINK *blst, int lev )
{
int		y, x;
int		ix, iy;
int		flag;
OBJECT	blob;
PIXEL	pix;

	DlinkInit( blst );
	
	for ( y = 0 ; y < info->ys ; y++ ) {

		for ( x = 0 ; x < info->xs ; x++ ) {
		
			if ( info->layer[ y ][ x ] != lev ) continue;
			if ( info->mark[ y ][ x ] ) continue;

			flag = 1;
			
			for ( iy = y-SCANSIZE ; iy <= y+SCANSIZE ; iy++ ) {
	
				if ( iy < 0 || iy >= info->ys ) continue;
					
				for ( ix = x-SCANSIZE ; ix <= x+SCANSIZE ; ix++ ) {
					
					if ( ix < 0 || ix >= info->xs ) continue;
					if ( iy == y && ix == x ) continue;
					
					if ( info->layer[ y ][ x ] < info->layer[ iy ][ ix ] &&
						 ! info->mark[ iy ][ ix ] ) {
					
						flag = 0;
						break;
					}
				}
				
				if ( ! flag ) break;
			}
			
			if ( flag ) {
			
				pix	   = MallocPixel();
				pix->y = y;
				pix->x = x;
				pix->v = info->layer[ y ][ x ];
								
				blob		 = MallocObject();
				blob->pixels = DLINK_EMPTY;
				blob->border = DLINK_EMPTY;
				blob->curr	 = pix;
				blob->id	 = info->blobid;
				blob->max	 = lev * 1.0;
				blob->min	 = lev * 0.3;
				blob->area	 = 1;
				blob->peri	 = 0;
				blob->atree	 = NULL;
				blob->ptree	 = NULL;
							
				DlinkAddTail( &(blob->pixels), pix );
				DlinkAddTail( blst, blob );
				
				info->mark[ y ][ x ] = info->blobid++;
				while ( blob->curr ) BlobPix( info, blob );
			}
		}
	}
}

/*
*	Removes invalid blobs and processes the rest on a list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	BlobBlob( INFO *info, DLINK *blst )
{
OBJECT	p, pp;

	for ( p = DlinkFirst( blst ) ; p ; p = pp ) {
		
		pp = DlinkNext( p );
		BlobTrim( info, p );
		
		if ( ! p->area ) {
		
			DlinkRemove( blst, p );
			Free( p );
		}
	}
}

/*
*	Processes the next pixel on a blob's list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	BlobPix( INFO *info, OBJECT blob )
{
int		x, y;
PIXEL	pix;
int		ix, iy;

	if ( ! blob->curr ) {
	
		return;
	}
	
	y = blob->curr->y;
	x = blob->curr->x;
	
	for ( iy = y-1 ; iy <= y+1 ; iy++ ) {
	
		if ( iy < 0 || iy >= info->ys ) continue;
		
		for ( ix = x-1 ; ix <= x+1 ; ix++ ) {
		
			if ( ix < 0 || ix >= info->xs ) continue;
			if ( iy == y && ix == x ) continue;
			if ( ix != x && iy != y ) continue;
			
			// We reject the pixel if:
			
			if ( info->mark[ iy ][ ix ] ) {
			
				continue;
			}
				
			if ( info->layer[ iy ][ ix ] > info->layer[ y ][ x ] ) {
			
				continue;
			}
			
			if ( info->layer[ iy ][ ix ] < info->min_alev ) {
			
				continue;
			}
			
			if ( info->layer[ iy ][ ix ] < blob->min ) {
			
//				continue;
			}
			
			if ( info->layer[ iy ][ ix ] > blob->max ) {
			
//				continue;
			}
			
			// Pixel accepted.
			
			pix = MallocPixel();
			pix->y = iy;
			pix->x = ix;
			pix->v = info->layer[ iy ][ ix ];
			info->mark[ iy ][ ix ] = blob->id;
			DlinkAddTail( &(blob->pixels), pix );
			blob->area++;
		}
	}			
	
	blob->curr = DlinkNext( blob->curr );
}

/*
*	Trims a blob
*	~~~~~~~~~~~~
*/

static	void	BlobTrim( INFO *info, OBJECT blob )
{
PIXEL	p;

	// Blobs that contain less than min_pixs pixels are rejected
	
	if ( blob->area < info->min_pixs ) {
	
		while (( p = DlinkFirst( &(blob->pixels) ))) BlobUnpix( info, blob, p );
		return;
	}
	
	// First, we try to grow the blob. For each pixel we check if
	// it has an adjacent empty pixel. If it has, then we check if 
	// the empty pixel has two neighbours on the same side
	// If yes, we add the pixel to the list. Do it until no more
	// is found.
	
	while ( BlobGrow( info, blob ) );
	
	// Now we delete all pixels which have only one neighbour
	
 	while ( BlobShrink( info, blob ) );
	
	// Blobs that contain less than min_pixs pixels are rejected again
	
	if ( blob->area < info->min_pixs ) {
	
		while ( (p=DlinkFirst( &(blob->pixels)) ) ) BlobUnpix( info, blob, p );
	}
}

/*
*	Removes the given pixel from the blob's list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	BlobUnpix( INFO *info, OBJECT blob, PIXEL p )
{
	DlinkRemove( &(blob->pixels), p );
	info->mark[ p->y ][ p->x ] = 0;
	Free( p );
	blob->area--;
}

/*
*	Returns the number of neighbours of a pixel
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	If ID is 0, then any occupied cell counts.
*	Cells out of image are occupied by a nonexistent ID
*/

static	int		BlobNeighbour( INFO *info, int y, int x, int id )
{
int		n;

	if ( x < 0 || y < 0 || x >= info->xs || y >= info->ys ) return( 0 );
	
	n = 0;
	n += ( y > 0 && info->mark[ y-1 ][ x ] == id );
	n += ( x > 0 && info->mark[ y ][ x-1 ] == id );
	n += ( y < info->ys-1 && info->mark[ y+1 ][ x ] == id );
	n += ( x < info->xs-1 && info->mark[ y ][ x+1 ] == id );
	
 	if ( id )
	
		return( n );
	else
		return( 4-n );
}

/*
*	Grows a blob
*	~~~~~~~~~~~~
*	Returns if it added a pixel or not
*/

static	int		BlobGrow( INFO *info, OBJECT blob )
{
int		flag;
PIXEL	p, pix;
int		y, x;
int		yy, xx;
int		i;
static	int	xoff[] = { -1, 0, 1, 0 };
static	int	yoff[] = { 0, -1, 0, 1 };

	flag = 0;
	
	for ( p = DlinkFirst( &(blob->pixels) ) ; p ; p = DlinkNext( p ) ) {
	
		y = p->y;
		x = p->x;
		
		if ( y < 2 || x < 2 || y > info->ys - 3 || x > info->xs - 3 ) continue;
		
		for ( i = 0 ; i < 4 ; i++ ) {
		
			yy = xoff[ i ] + y;
			xx = yoff[ i ] + x;
			
			if ( info->mark[ yy ][ xx ] ) continue;
			
			if ( ( info->mark[ yy-1 ][ xx ] == blob->id &&
				   info->mark[ yy+1 ][ xx ] == blob->id ) ||
				 ( info->mark[ yy ][ xx-1 ] == blob->id &&
				   info->mark[ yy ][ xx+1 ] == blob->id ) ) {
				
				pix = MallocPixel();
				pix->y = yy;
				pix->x = xx;
				pix->v = info->layer[ yy ][ xx ];
				info->mark[ yy ][ xx ] = blob->id;
				DlinkAddTail( &(blob->pixels), pix );
				blob->area++;
				flag = 1;
			}
		}
	}
	
	return( flag );
}

/*
*	Shrinks a blob
*	~~~~~~~~~~~~~~
*	Returns if it removed a pixel
*/

static	int		BlobShrink( INFO *info, OBJECT blob ) 
{
int		flag;
PIXEL	p, pp;
int		y, x;
int		yy, xx;
int		i;

	flag = 0;
	
	for ( p = DlinkFirst( &(blob->pixels) ) ; p ; p = pp ) {
	
		pp = DlinkNext( p );
		y  = p->y;
		x  = p->x;
		
		// We remove pixels that are too close to the edges
		
		if ( y < 1 || x < 1 || y > info->ys - 2 || x > info->xs - 2 ) {
		
			BlobUnpix( info, blob, p );
			flag = 1;
			continue;
		}
		
		// We remove pixels that have only one neighbour
		
		if ( BlobNeighbour( info, p->y, p->x, blob->id ) < 2 ) {
		
			BlobUnpix( info, blob, p );
			flag = 1;
			continue;
		}
	}
	
	return( flag );						
}

/****************************************************************************/
/*								Twig search									*/
/****************************************************************************/

static	void	TwigSearch( INFO *info )
{
int		x, y;
OBJECT	twig;
DLINK	temp;
int		size;
	
	// Select the pixels that are bright enough to start a grow,
	// then grow the object
	
	for ( y = 0 ; y < info->ys ; y++ ) {
	
		if ( info->progress ) ProgressUpdate( y, info->ys );
		
		for ( x = 0 ; x < info->xs ; x++ ) {
		
			if ( info->mark[ y ][ x ] ) continue;
			if ( info->layer[ y ][ x ] < info->min_slev ) continue;
			
			// The pixel is bright enough to start and it is not taken yet
			// Let's try to grow a twig.
			
			if ( ! ( size = TwigGrow( info, &temp, y, x ) ) ) continue;
					
			// We have a new twig. Now temp is the pixel list.
			// Create a new object object and put all that we
			// can in it, then add it to the list of objects.
			
			twig		 = MallocObject();
			twig->border = DLINK_EMPTY;			
			twig->pixels = temp;
			twig->area   = size;
			twig->peri	 = 0;
			twig->id	 = info->blobid++;
			twig->atree	 = NULL;
			twig->ptree	 = NULL;

			DlinkAddTail( info->list, twig );
		}
	}
}	
	
/*
*	From a given starting point it tries to grow a twig
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Return the number of pixels within the twig
*/

static	int		TwigGrow( INFO *info, DLINK *twig, int y, int x )
{
PIXEL	wrk;
PIXEL	tmp;
int		xx, yy;
int		size;
int		i;
DLINK	wqu;
static	int	yoff[] = { -1, -1, -1,  0,  0, +1, +1, +1 };
static	int	xoff[] = { -1,  0, +1, -1, +1, -1,  0, +1 };

	
	// Get a pixel and fill it in with the starting point
	
	DlinkInit( twig );
	DlinkInit( &wqu );
	
	wrk	   = MallocPixel();
	wrk->y = y;
	wrk->x = x;
	
	DlinkAddTail( twig, wrk );
	
	wrk	   = MallocPixel();
	wrk->y = y;
	wrk->x = x;
	
	DlinkAddTail( &wqu, wrk );
	
	info->mark[ y ][ x ] = TRUE;
	size = 1;
		
	while ( ( wrk = DlinkRemoveFirst( &wqu ) ) ) {
	
		// Get the current pixel
		
		y = wrk->y;
		x = wrk->x;
		
		// Check the 8 adjacent pixels and try to add them
		// to the object
		
		for ( i = 0 ; i < 8 ; i++ ) {
		
			yy = yoff[ i ] + y;
			xx = xoff[ i ] + x;
			
			if ( yy < 0 || yy >= info->ys ) continue;
			if ( xx < 0 || xx >= info->xs ) continue;
			if ( info->mark[ yy ][ xx ]   ) continue;
			
			if ( info->layer[ yy ][ xx ] < info->min_alev ) continue;
			
			// Pixel is accepted, add it to the object
			
			tmp	   = MallocPixel();
			tmp->y = yy;
			tmp->x = xx;
			tmp->v = info->layer[ yy ][ xx ];
			
			DlinkAddTail( twig, tmp );
			
			// Also add to the working queue
			
			tmp = MallocPixel();
			tmp->y = yy;
			tmp->x = xx;
			
			DlinkAddTail( &wqu, tmp );
			
			info->mark[ yy ][ xx ] = TRUE;
			size++;
		}
		
		Free( wrk );		
	}
	
	// Return the size of the twig
	
	return( size );
}

/****************************************************************************/
/*							Round object search								*/
/****************************************************************************/

/*
*	Based on elliptic search
*	~~~~~~~~~~~~~~~~~~~~~~~~
*	Builds large objects by finding blobs that are adjacent.
*	These are merged into one large object. The center of adjacent blobs
*	will be connected. The full connection is then trimmed by finding
*	the minimal cost spanning tree. Thus, small round objects are
*	reduced to point, elongated type objects are reduced to a tree.
*/

