/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	2D Fourier transform based image filter									*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~									*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <setjmp.h>
#include <math.h>

#include <types.h>
#include <image.h>
#include <memory.h>
#include <progress.h>
#include <process.h>

/****************************************************************************/
/*						Local definitions									*/
/****************************************************************************/

#define	PI	3.141562654

typedef	struct	{

	float	re;
	float	im;

} CPX;

typedef	struct	{

	int		M;						/* M so that 2^M >= image_max_size		*/
	int		N;						/* 2^M									*/
	float	*window;				/* Smoothing window coefficients		*/
	CPX		*twiddle;				/* Rotator vectors for FFT				*/
	CPX		**kernel;				/* Filter kernel						*/
	CPX		**data;					/* Image data							*/
	int		runs;					/* How many runs of FFT					*/
	int		done;					/* How many is done						*/
		
} FILTER;

/****************************************************************************/
/*						Prototypes											*/
/****************************************************************************/

static	void	FilterFilter( FILTER *filter, IMAGE *result, IMAGE *image );
static	void	FilterTwiddleCreate( FILTER *filt );
static	void	FilterKernelCreate( FILTER *filt, double radius );
static	void	FilterFFT2D( FILTER *filt, CPX **data, int inv ); 
static	void	FilterFFT( FILTER *filt, CPX *data, int inv );
static	double	FilterBlackmanHarris( int n, int N );
static	void	FilterTranspose( FILTER *filt, CPX **array );
static	void	FilterLoadImage( FILTER *filt, IMAGE *img );
static	void	FilterSaveImage( FILTER *filt, IMAGE *img, CPX **data );
static	void	FilterMultiply( FILTER *filt, CPX **a, CPX **b );

/****************************************************************************/
/*						Public functions									*/
/****************************************************************************/

/*
*	Filters the R and G layers on the image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*
*		old:		Original image
*		radius:		The radius of the filter kernel
*	
*	Return value
*
*		A pointer to the new image or NULL if could not finish
*
*	Receives a 2-layer image. For both layers, applies a convolution
*	filter with a raised cosine kernel of the given radius.
*	Returns the resulting image or NULL, if the processing was aborted.
*/

IMAGE	*FilterImage( IMAGE *image, double radius )
{
int		s, n, m;
FILTER	*filter;
IMAGE	*result;
jmp_buf	state;

	// If radius is less than 0.5 pixel, do nothing
	
	if ( radius < 0.5 ) return( ImageDup( image ) );
			
	// Create the result image
	
	result = ImageCreate( image->y, image->x, 2 );
	
	// Work out the size of the enclosing 2^M by 2^M square
	
	s = image->x > image->y ? image->x : image->y;
	for ( n = 1, m = 0 ; n < s ; n <<= 1, m++ );
	
	// Get all the memory that we will need and fill basic parameters
	// Filter runs is how many times we call the 1D FFT. A 2D FFT calls
	// the 1D one 2N times. We have a normal 2D FFT for the filter and
	// a normal and an inverse for the image -> ( 1 + 2 ) * 2 * N
	
	filter			= Malloc( sizeof( FILTER ) );
	filter->kernel	= MallocArray( n, n, sizeof( CPX ) );
	filter->data	= MallocArray( n, n, sizeof( CPX ) );
	filter->window	= Malloc( n * sizeof( float ) );
	filter->twiddle = Malloc( n * sizeof( CPX ) );
	filter->N		= n;
	filter->M		= m;
	filter->runs	= n * 2 * ( 2 + 1 );
	filter->done	= 0;
		
	
	if ( setjmp( state ) ) {
	
		// Aborted, release everything and report failure
		
		Free( filter->data    );
		Free( filter->kernel  );
		Free( filter->window  );
		Free( filter->twiddle );
		Free( filter );
		ImageDestroy( result );
		return( NULL );
	}
	
	ProgressInit( "Filtering image", state );
	
	// Initialise the rotator vector table
	
	FilterTwiddleCreate( filter );
	
	// Create the filter kernel
	
	FilterKernelCreate( filter, radius );
	
	// Process the red and the green data layers
	
	FilterFilter( filter, result, image );
	
	// Finished
	
	ProgressDone();
	
	// Release resources
	
	Free( filter->data    );
	Free( filter->kernel  );
	Free( filter->window  );
	Free( filter->twiddle );
	Free( filter );

	// Done
	
	return( result );
}

/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/

/*
*	Processes the R and G layers of an image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Since all operations are linear and the data is all real, we can save 
*	half of the transforms by using the real and imaginary parts separately.
*/

static	void	FilterFilter( FILTER *filter, IMAGE *result, IMAGE *image )
{
	FilterLoadImage( filter, image );
	FilterFFT2D( filter, filter->data, FALSE );
	FilterMultiply( filter, filter->data, filter->kernel );
	FilterFFT2D( filter, filter->data, TRUE );
	FilterSaveImage( filter, result, filter->data );
}

/*
*	Generates the raised cosine kernel
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	It receives the radius of the kernel and the size of the image
*	matrix. It returns a matrix that has to be piece-wise multiplied
*	with the Fourier transform of the image. To avoid Gibbs oscillation
*	artefacts, a Blackman-Harris windowing function is also applied on
*	the filter kernel. The returned kernel is uniform, it needs no
*	correction when the inverse FFT is done.
*	The funny index calculation on the windowing vector is because of
*	the way the FFT places the result coefficients into the array (see 
*	comment at FFT).
*/

static	void	FilterKernelCreate( FILTER *filt, double radius )
{
double	sum, t;
int		N;
int		y, x;

	N = filt->N;
	
	for ( x = 0 ; x < N ; x++ ) {
	
		filt->window[ ( x + N/2 ) % N ] = sqrt( FilterBlackmanHarris( x, N ) );
	}

	for ( sum = 0, y = 0 ; y <= radius ; y++ ) {
	
		for ( x = 0 ; x <= radius ; x++ ) {
		
			t	 = sqrt( y*y + x*x ) / radius * PI;
			t	 = t > PI ? 0 : ( cos( t ) + 1 );	
			
			filt->kernel[ ( N + y ) % N ][ ( N + x ) % N ].re = t;
			filt->kernel[ ( N + y ) % N ][ ( N - x ) % N ].re = t;
			filt->kernel[ ( N - y ) % N ][ ( N + x ) % N ].re = t;
			filt->kernel[ ( N - y ) % N ][ ( N - x ) % N ].re = t;
			
			if ( ! y ) t /= 2;
			if ( ! x ) t /= 2;
			
			sum += 4 * t;
			
			
		}
	}
	
	FilterFFT2D( filt, filt->kernel, FALSE );
	
	for ( y = 0 ; y < N ; y++ ) {
	
		for ( x = 0 ; x < N ; x++ ) {
	
			filt->kernel[ y ][ x ].re *= filt->window[y]*filt->window[x] / sum;
			filt->kernel[ y ][ x ].im = 0;
		}
	}
}

/*
*	Two dimensional Fourier transform
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Receives a pointer to an array of pointers to arrays of CPX.
*	The array is 2^M by 2^M large.
*	If 'inv' is true, then it is an inverse transform.
*
*	The DC coefficient will be in the [0][0] rather that [N/2][N/2],
*	due to the way the 1D FFT works!!!
*/

static	void	FilterFFT2D( FILTER *filt, CPX **data, int inv ) 
{
int		i;

	for ( i = 0 ; i < filt->N ; i++ ) FilterFFT( filt, data[ i ], inv );
	FilterTranspose( filt, data );
	for ( i = 0 ; i < filt->N ; i++ ) FilterFFT( filt, data[ i ], inv );
}

/*
*	Fast Fourier transform - straightforward radix-2 algorithm
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Transforms the data[] array, that has to be 2^M elements.
*	If 'inv' is TRUE, then it is an inverse transform.
*
*	Note that the resulting frequency coefficients will be
*	in the form of
*		0, 1, 2, ... N/2-1, -N/2, -N/2+1, ... -1
*	and not the natural
*		-N/2, -N/2+1, ... -1, 0, 1, 2, ... N/2-1
*	sequence.
*	This must be taken into account when manipulating the array in the
*	frequency domain.
*/

static	void	FilterFFT( FILTER *filt, CPX *data, int inv )
{
int		i, j, k, l, n1, n2, i1, i2;
CPX		t, tw;

	if ( ! ( ++filt->done & 63 ) ) ProgressUpdate( filt->done, filt->runs );
	
	n2 = filt->N;
	
	for ( i = 1 ; i <= filt->M ; i++ ) {

		n1 = n2;
		n2 >>= 1;
		i1 = 0;
		i2 = filt->N / n1;

		for ( j = 0 ; j < n2 ; j++ ) {

			tw.re = filt->twiddle[ i1 ].re;
			tw.im = inv ? filt->twiddle[ i1 ].im : -filt->twiddle[ i1 ].im;
			i1 += i2;
			
			for ( k = j ; k < filt->N ; k += n1 ) {

				l = k + n2;

				t.re = data[ k ].re - data[ l ].re;
				t.im = data[ k ].im - data[ l ].im;

				data[ k ].re += data[ l ].re;
				data[ k ].im += data[ l ].im;
				data[ l ].re  = t.re * tw.re + t.im * tw.im;
				data[ l ].im  = t.im * tw.re - t.re * tw.im;
			}
		}
	}

	if ( inv ) {

		for ( i = 0 ; i < filt->N ; i++ ) {

			data[ i ].re /= filt->N;
			data[ i ].im /= filt->N;
		}
	}

	for ( j = 0, i = 0 ; i < filt->N-1 ; i++ ) {

		if ( i < j ) {
			
			t.re = data[ j ].re;
			t.im = data[ j ].im;

			data[ j ].re = data[ i ].re;
			data[ j ].im = data[ i ].im;

			data[ i ].re = t.re;
			data[ i ].im = t.im;
		}

		for ( k = filt->N / 2 ; k <= j ; k >>= 1 )

			j -= k;

		j += k;
	}
}

/*
*	Creates the twiddle factor table
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void		FilterTwiddleCreate( FILTER *filt )
{
int		i;
int		N;

	N = filt->N;
		
	for ( i = 0 ; i < N ; i++ ) {

		filt->twiddle[ i ].im = sin( 2 * PI / N * i );
		filt->twiddle[ i ].re = cos( 2 * PI / N * i );
	}
}

/*
*	Blackman - Harris window function
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Calculates the next coefficient of the optimal
*	Blackman-Harris 4 term ( sidelobe < -92 dB ) window
*/

static	double	FilterBlackmanHarris( int n, int N )
{
double	w, t;	
int		m;	
double	temp;
static	double	a[ 4 ] = { 0.35875, 0.48829, 0.14128, 0.01168 };
	
	for ( w = 0.0, m = 0 ; m < 4 ; m++ ) {

		t  = a[ m ] * cos( 2 * PI * m * n / (N-1) );
		w += m & 1 ? -t : t;
	}

	return( w );
}

/*
*	Transposes a matrix
*	~~~~~~~~~~~~~~~~~~~
*	Receives an N by N complex matrix and transposes it
*/

static	void	FilterTranspose( FILTER *filt, CPX **array )
{
int		y, x;
CPX		t;

	for ( y = 0 ; y < filt->N ; y++ ) {
	
		for ( x = y + 1 ; x < filt->N ; x++ ) {
		
			t = array[ y ][ x ];
			array[ y ][ x ] = array[ x ][ y ];
			array[ x ][ y ] = t;
		}
	}
}

/*
*	Loads a layer of an image into a complex array
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FilterLoadImage( FILTER *filt, IMAGE *img )
{
int		N;
int		xo, yo;
int		x, y;

	N = filt->N;
	
	yo = ( N - img->y ) / 2;
	xo = ( N - img->x ) / 2;
	
	for ( y = 0 ; y < img->y ; y++ ) {
	
		for ( x = 0 ; x < img->x ; x++ ) {
		
			filt->data[ y+yo ][ x+xo ].re = img->im[ R ][ y ][ x ];
			filt->data[ y+yo ][ x+xo ].im = img->im[ G ][ y ][ x ];
		}
	}
}

/*
*	Saves the real part of a compex array into an img
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FilterSaveImage( FILTER *filt, IMAGE *img, CPX **data )
{
int		N;
int		xo, yo;
int		x, y;
int		d;

	N = filt->N;
	
	yo = ( N - img->y ) / 2;
	xo = ( N - img->x ) / 2;
	
	for ( y = 0 ; y < img->y ; y++ ) {
	
		for ( x = 0 ; x < img->x ; x++ ) {
		
			d = data[ y+yo ][ x+xo ].re;
			if ( d <   0 ) d =   0;
			if ( d > 255 ) d = 255;
			img->im[ R ][ y ][ x ] = d;
			
			d = data[ y+yo ][ x+xo ].im;
			if ( d <   0 ) d =   0;
			if ( d > 255 ) d = 255;
			img->im[ G ][ y ][ x ] = d;
		}
	}
}

/*
*	Multiplies a complex array piecewise with one that has only reals
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FilterMultiply( FILTER *filt, CPX **a, CPX **b )
{
int		x, y;

	for ( y = 0 ; y < filt->N ; y++ ) {
	
		for ( x = 0 ; x < filt->N ; x++ ) {
		
			a[ y ][ x ].re *= b[ y ][ x ].re;
			a[ y ][ x ].im *= b[ y ][ x ].re;
		}
	}
}
