/****************************************************************************/
/* Copyright 2004, � Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Object extraction - fixup procedures									*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~									*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zolt�n K�csi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <setjmp.h>

#include <types.h>
#include <memory.h>
#include <dlinklist.h>
#include <objects.h>
#include <process.h>
#include <progress.h>
#include "process_intern.h"

/****************************************************************************/
/*						Local definitions									*/
/****************************************************************************/

typedef	struct {

	int		ys, xs;					/* Map size								*/
	int		**mark;					/* The pixel map						*/
	int		nextid;					/* Highest ID we found					*/
	int		objcnt;					/* Number of objects on the list		*/
	int		runcnt;					/* Process count						*/
	DLINK	list;					/* Object list							*/
		
} INFO;
	
#define	MARK_FLAG	0x80000000

#define	OFFSET		4
#define	BORD_THOLD	10

/****************************************************************************/
/*						Prototypes											*/
/****************************************************************************/

static	void	FixupMarkOriginal( INFO *info );
static	void	FixupRemoveFromMark( INFO *info, OBJECT obj );
static	void	FixupFindPerimeter( INFO *info, OBJECT obj );
static	void	FixupSlitPerimeter( INFO *info, OBJECT obj );
static	void	FixupTrimPerimeter( INFO *info, OBJECT obj );
static	OBJECT	FixupCheckDisjunct( INFO *info, OBJECT obj );
static	void	FixupFillHoles( INFO *info, OBJECT obj );
static	void	FixupFinalise( INFO *info, OBJECT obj );
static	void	FixupCalcSize( INFO *info, OBJECT obj );
static	int		FixupIsBoundary( INFO *info, OBJECT obj );

/****************************************************************************/
/*						Public functions									*/
/****************************************************************************/

/*
*	Trims all objects on a list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		objs		A pointer to a DLINK containing the objects to trim
*		y, x		The size of the image
*		txt			Text for the progress bar
*		size		Minimal size (in pixels) for an object
*		fill		Flag indicating if holes should be filled
*		excl		Flag indicating if boundary objects should be discarded
*
*	Return value
*		The length of the final queue or -1 if it was aborted
*
*	The returned list is sorted by object area, from largest to smallest.
*/

int		ObjectFixup( DLINK *objs, int y, int x, char *txt, 
					 int size, int fill, int excl )
{
INFO	info;
OBJECT	obj, new;
int		objnum;
jmp_buf	state;

	info.ys		= y;
	info.xs		= x;
	info.mark	= MallocArray( info.ys+2*OFFSET, 
							   info.xs+2*OFFSET, sizeof( int ) );
	info.list	= *objs;
	info.runcnt = 0;
	DlinkInit( objs );
		
	if ( setjmp( state ) ) {
	
		// Aborted. Destroy everything and report failure
		
		if ( info.mark ) Free( info.mark );
		ObjectDestroyList( objs );
		ObjectDestroyList( &info.list );
		return( -1 );
	}		
	
	objnum = 0;
	ProgressInit( txt, state );
	
	// Draw all objects into the mark and offset pixels
	
	FixupMarkOriginal( &info );
	info.objcnt *= 3;
	
	// For every object do the trimming
	
	for ( obj = DlinkFirst( &info.list ) ; obj ; ) {
	
		ProgressUpdate( info.runcnt++, info.objcnt );
		FixupSlitPerimeter( &info, obj );
		FixupTrimPerimeter( &info, obj );
		
		if ( ( new = FixupCheckDisjunct( &info, obj ) ) ) {
		
			info.objcnt += 3;
			DlinkAddBefore( &info.list, new, obj );
			obj = new;
		}
		else {
		
			obj = DlinkNext( obj );
		}	
	}
	
	// Cull objects that are too small or that are on a boundary and we
	// have to discard those
	
	while ( ( obj = DlinkRemoveFirst( &info.list ) ) ) {
	
		ProgressUpdate( info.runcnt++, info.objcnt );
		FixupCalcSize( &info, obj );
		
		if ( obj->area < size || ( excl && FixupIsBoundary( &info, obj ) ) ) {
		
			FixupRemoveFromMark( &info, obj );
			ObjectDestroy( obj );
		}
		else {
		
			DlinkAddTail( objs, obj );
		}
	}
	
	// Move the objects back to the local list and do the final processing
	
	DlinkCat( &info.list, objs );
	
	while ( ( obj = DlinkRemoveFirst( &info.list ) ) ) {
	
		ProgressUpdate( info.runcnt++, info.objcnt );
		if ( fill ) FixupFillHoles( &info, obj );
		FixupFinalise( &info, obj );
		obj->id = ++objnum;
		
		// We keep the list sorted by size, in descending order
		
		for ( new = DlinkFirst( objs ) ; new ; new = DlinkNext( new ) ) {
			
			if ( obj->area >= new->area ) {
				
				DlinkAddBefore( objs, obj, new );
				break;
			}
		}
		
		if ( ! new ) DlinkAddTail( objs, obj );
	}

	Free( info.mark );
	info.mark = NULL;
	ProgressDone();
	return( objnum );
}

/*
*	Calculates the perimeters and the spatial trees for a list
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters
*		objs		A pointer to a DLINK containing the objects to trim
*		y, x		The size of the image
*
*	Return value
*		The length of the final queue
*
*	This function does not change the objects whatsoever. 
*	It merely calculates the borders and the creates distance division trees.
*/

int		ObjectQuickfix( DLINK *objs, int y, int x )
{
INFO	info;
OBJECT	obj;
int		objnum;

	info.ys		= y;
	info.xs		= x;
	info.mark	= MallocArray( info.ys+2*OFFSET, 
							   info.xs+2*OFFSET, sizeof( int ) );
	info.list	= *objs;
	info.runcnt = 0;
	DlinkInit( objs );
		
	objnum = 0;
	
	// Draw all objects into the mark and offset pixels
	
	FixupMarkOriginal( &info );
	
	while ( ( obj = DlinkRemoveFirst( &info.list ) ) ) {
	
		FixupCalcSize( &info, obj );
		DlinkAddTail( objs, obj );
	}
	
	// Move the objects back to the local list and do the final processing
	
	DlinkCat( &info.list, objs );
	
	while ( ( obj = DlinkRemoveFirst( &info.list ) ) ) {
	
		FixupFinalise( &info, obj );
		obj->id = ++objnum;
		DlinkAddTail( objs, obj );
	}

	Free( info.mark );
	info.mark = NULL;
	return( objnum );
}


/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/
	
/*
*	Fills the mark with the current objects
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FixupMarkOriginal( INFO *info )
{
OBJECT	f;
PIXEL	p;

	info->nextid = 1;
	info->objcnt = 0;
	
	for ( f = DlinkFirst( &info->list ) ; f ; f = DlinkNext( f ) ) {
	
		info->objcnt++;
		
		f->id = info->nextid++;
				
		for ( p = DlinkFirst( &f->pixels ) ; p ; p = DlinkNext( p ) ) {
		
			// Offset pixels so that they are at least 4 pixels from the
			// mark's edge => we don't have to check boundary conditions.
			
			p->y += OFFSET;
			p->x += OFFSET;
			
			info->mark[ p->y ][ p->x ] = f->id;
		}
	}
}

/*
*	Deletes the given object from the mark
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FixupRemoveFromMark( INFO *info, OBJECT obj )
{
PIXEL	pix;

	for ( pix = DlinkFirst( &obj->pixels ) ; pix ; pix = DlinkNext( pix ) ) {
	
		info->mark[ pix->y ][ pix->x ] = 0;
	}
}
	
/*
*	Calculates the raw pixel count in an object
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FixupCalcSize( INFO *info, OBJECT obj )
{
PIXEL	p;

	DlinkCat( &obj->pixels, &obj->border );
	
	obj->area = 0;
	obj->peri = 0;

	for ( p = DlinkFirst( &obj->pixels ) ; p ; p = DlinkNext( p ) ) obj->area++;
}

/*
*	Identifies the perimeter pixels of an object
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Removes them from the pixels list and places on the border list
*/

static	void	FixupFindPerimeter( INFO *info, OBJECT obj )
{
PIXEL	p;
DLINK	tmp;
int		cnt, dy, dx;

	DlinkInit( &tmp );
	
	obj->area = 0;
	obj->peri = 0;
	
	// If there were leftover perimeter pixels, give them back to
	// the pixel pool
	
	DlinkCat( &obj->pixels, &obj->border );
	
	// Identify all pixels that have a neighbour that does not belong to
	// this object
	
	while ( ( p = DlinkRemoveFirst( &obj->pixels ) ) ) {
	
		if ( ! info->mark[ p->y ][ p->x ] ) {
		
			// Dud pixel, get rid of it
			
			Free( p );
			continue;
		}
		
		cnt = 0;
		cnt += ( info->mark[ p->y + 0 ][ p->x + 1 ] == obj->id );
		cnt += ( info->mark[ p->y + 0 ][ p->x - 1 ] == obj->id );
		cnt += ( info->mark[ p->y + 1 ][ p->x + 0 ] == obj->id );
		cnt += ( info->mark[ p->y - 1 ][ p->x + 0 ] == obj->id );

		
		if ( cnt == 4 ) {
		
			DlinkAddTail( &tmp, p );
			obj->area++;
		}
		else {
		
			DlinkAddTail( &obj->border, p );
			obj->peri++;
		}
	}
	
	obj->pixels = tmp;
	obj->area  += obj->peri;
}

/*
*	Removes pixels which are closer than 2 pixels from an other object
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	The perimeter pixels are removed from the border queue and placed
*	back onto the pixels queue. Deleted pixels are removed from the mark
*	as well.
*/

static	void	FixupSlitPerimeter( INFO *info, OBJECT obj )
{
PIXEL	p;
DLINK	tmp;
int		mark, flag, dy, dx;

	for ( mark = TRUE ; mark ; mark = ! mark ) {
	
		FixupFindPerimeter( info, obj );
		
		DlinkInit( &tmp );
	
		while ( ( p = DlinkRemoveFirst( &obj->border ) ) ) {
	
			if ( ! info->mark[ p->y ][ p->x ] ) {
		
				// Pixel was deleted, discard it
			
				Free( p );
				continue;
			}
		
			flag = FALSE;
		
			for ( dy = -1 ; ! flag && dy <= 1 ; dy++ ) {
		
				for ( dx = -1 ; ! flag && dx <= 1 ; dx++ ) {
			
					if ( ! info->mark[ p->y + dy ][ p->x + dx ] ) {
				
						// Empty pixel, don't worry about it
					
						continue;
					}

					if ( info->mark[ p->y + dy ][ p->x + dx ] != obj->id ) {
				
						// Foreign pixel, make note that we should delete
						// the current border pixel
					
						flag = TRUE;
					}
				}
			}
		
			if ( flag ) {
		
				// Delete current pixel
			
				info->mark[ p->y ][ p->x ] = 0;
				mark = FALSE;
				Free( p );
			}
			else {
		
				// Keep pixel
			
				DlinkAddTail( &tmp, p );
			}
		}
		
		// Put kept pixels back to the pixel list, we will need to
		// recalculate borders anyway
	
		DlinkCat( &obj->pixels, &tmp );
	}
}

/*
*	Removes perimeter pixels that have only perimeter neighbours
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	This process can cut an object into two (or more), so we have 
*	to test for that after this call.
*	Removed pixels are deleted from the mark. Border pixels are
*	placed back to the pixels queue.
*/

static	void	FixupTrimPerimeter( INFO *info, OBJECT obj )
{
PIXEL	p;
DLINK	keep, drop;
int		flag, dy, dx, mark;

static int	oy[] = { -1,  0,  0, +1 };
static int	ox[] = {  0, -1, +1,  0 };
		
	for ( ;; ) {
	
		DlinkInit( &keep );
		DlinkInit( &drop );
	
		// Calculate perimeter
		
		FixupFindPerimeter( info, obj );
		
		// Mark perimeter pixels on mark
	
		for ( p = DlinkFirst( &obj->border ) ; p ; p = DlinkNext( p ) ) {
	
			info->mark[ p->y ][ p->x ] |= MARK_FLAG;
		}
	
		// First, remove all perimeter pixels that have no 
		// connection to the object's internals

		flag = FALSE;
		
		while ( ( p = DlinkRemoveFirst( &obj->border ) ) ) {
	
			for ( mark = FALSE, dy = -1 ; ! mark && dy <= 1 ; dy++ ) {
		
				for ( dx = -1 ; ! mark && dx <= 1 ; dx++ ) {
			
					if ( info->mark[ p->y + dy ][ p->x + dx ] == obj->id ) {
					
						mark = TRUE;
					}
				}
			}
			
			// If mark is false, then this pixel should be deleted
			
			if ( ! mark ) {
			
				info->mark[ p->y ][ p->x ] = 0;
				flag = TRUE;
				Free( p );
			}
			else {
			
				DlinkAddTail( &keep, p );
			}
		}
		
		if ( flag ) {
		
			// We deleted border pixels, do it again
			// Before starting again, we unmark the kept pixels 
			// and put them back to the pixel list
			
			while ( ( p = DlinkRemoveFirst( &keep ) ) ) {
		
				info->mark[ p->y ][ p->x ] &= ~MARK_FLAG;
				DlinkAddTail( &obj->pixels, p );
			}
		
			continue;
		}
		else {
		
			// We've done nothing, place the kept list back to 
			// the object's perimeter list
			
			obj->border = keep;
			DlinkInit( &keep );
		}
				
		// Now get rid of the corner pixels
		
		while ( ( p = DlinkRemoveFirst( &obj->border ) ) ) {
		
			flag = FALSE;
			
			if ( info->mark[ p->y + oy[ 0 ] ][ p->x + ox[ 0 ] ] & MARK_FLAG &&
			     info->mark[ p->y + oy[ 1 ] ][ p->x + ox[ 1 ] ] & MARK_FLAG ) {
						
				flag = TRUE;
			}
			
			if ( info->mark[ p->y + oy[ 0 ] ][ p->x + ox[ 0 ] ] & MARK_FLAG &&
			     info->mark[ p->y + oy[ 2 ] ][ p->x + ox[ 2 ] ] & MARK_FLAG ) {
						
				flag = TRUE;
			}
			
			if ( info->mark[ p->y + oy[ 3 ] ][ p->x + ox[ 3 ] ] & MARK_FLAG &&
			     info->mark[ p->y + oy[ 1 ] ][ p->x + ox[ 1 ] ] & MARK_FLAG ) {
						
				flag = TRUE;
			}
			
			if ( info->mark[ p->y + oy[ 3 ] ][ p->x + ox[ 3 ] ] & MARK_FLAG &&
			     info->mark[ p->y + oy[ 2 ] ][ p->x + ox[ 2 ] ] & MARK_FLAG ) {
						
				flag = TRUE;
			}
			
			if ( flag ) {
			
				// This is a corner pixel, delete it
				
				info->mark[ p->y ][ p->x ] = 0;
				Free( p );
			}
			else {
			
				// Not a corner, keep it
				
				DlinkAddTail( &keep, p );
			}
		}
		
		// Kept pixels back to the border list
		
		obj->border = keep;
		DlinkInit( &keep );	
				
		while ( ( p = DlinkRemoveFirst( &obj->border ) ) ) {
	
			flag = 0;
		
			for ( dy = -1 ; dy <= 1 ; dy++ ) {
		
				for ( dx = -1 ; dx <= 1 ; dx++ ) {
			
					if ( info->mark[ p->y + dy ][ p->x + dx ] & MARK_FLAG ) {
				
						flag++;
						continue;
				
					}
				}
			}
		
			// If flag is != 3, then this is not a pixel with exactly
			// two neighbours, so we delete it.
		
			DlinkAddTail( flag == 3 ? &keep : &drop, p );
		}
	
		// Now, we run through both lists and unflag the pixels and
		// remove them, if necessary. Since new border calculation is needed,
		// the kept pixels will be placed back to the object area list
		// Make note if we actually deleted a pixel or not
			
		while ( ( p = DlinkRemoveFirst( &keep ) ) ) {
	
			info->mark[ p->y ][ p->x ] &= ~MARK_FLAG;
			DlinkAddTail( &obj->pixels, p );
		}
	
		flag = FALSE;
	
		while ( ( p = DlinkRemoveFirst( &drop ) ) ) {
	
			flag = TRUE;
			info->mark[ p->y ][ p->x ] = 0;
			Free( p );
		}
	
		// If there was no deletion, we are done
		
		if ( ! flag ) break;
	}
}

/*
*	This routine checks if an object contains disjunct parts
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	If it finds disjunct parts, it takes one of them and creates
*	a new object for that. The function returns the new object and
*	keeps the rest as the old one. When the old one is one single
*	piece, then it returns NULL
*/

static	OBJECT	FixupCheckDisjunct( INFO *info, OBJECT obj )
{
OBJECT	new;
DLINK	temp, sepa;
PIXEL	p, np;
int		dy, dx;

	DlinkInit( &temp );
	DlinkInit( &sepa );
	
	if ( ! ( p  = DlinkFirst( &obj->pixels ) ) ) return( NULL );
	np = MallocPixel();
	
	np->x = p->x;
	np->y = p->y;
	
	DlinkAddHead( &temp, np );
	
	while ( ( p = DlinkRemoveFirst( &temp ) ) ) {
	
		if ( info->mark[ p->y ][ p->x ] & MARK_FLAG ) {
		
			// Already been here
			
			Free( p );
			continue;
		}
		
		info->mark[ p->y ][ p->x ] |= MARK_FLAG;
		
		for ( dy = -1 ; dy <= 1 ; dy++ ) {
		
			for ( dx = -1 ; dx <= 1 ; dx++ ) {
			
				if ( ! info->mark[ p->y + dy ][ p->x + dx ] ) 
				
					continue;
					
				if ( info->mark[ p->y + dy ][ p->x + dx ] & MARK_FLAG )
				
					continue;
					
				if ( info->mark[ p->y + dy ][ p->x + dx ] != obj->id )
				
					continue;
					
				np = MallocPixel();
				np->y = p->y + dy;
				np->x = p->x + dx;
				DlinkAddTail( &temp, np );
			}
		}
		
		Free( p );
	}
	
	// Now temp is empty and all pixels, starting with the first pixel
	// of the object are marked by MARK_FLAG. Go through the object's
	// list and see if there are any pixels that are not marked.
	
	while ( ( p = DlinkRemoveFirst( &obj->pixels ) ) ) {
	
		if ( info->mark[ p->y ][ p->x ] & MARK_FLAG )
		
			DlinkAddTail( &sepa, p );
		else
			DlinkAddTail( &temp, p );
	}
	
	// Now all pixels of the object that were marked are on the sepa
	// list, the rest on the temp queue. If the temp queue is not empty,
	// then we have to create a new object
	
	if ( DlinkFirst( &temp ) ) {
	
		// We need a new object and have to change the marked pixels
		// to the new object ID
		
		new = MallocObject();
		new->id = info->nextid++;
		DlinkInit( &new->pixels );
		DlinkInit( &new->border );
		new->atree = NULL;
		new->ptree = NULL;
		
		while ( ( p = DlinkRemoveFirst( &sepa ) ) ) {
		
			info->mark[ p->y ][ p->x ] = new->id;
			DlinkAddTail( &new->pixels, p );
		}
		
		// Put the rest of the pixels back to the original object
		
		obj->pixels = temp;
		
		// Return the new object
		
		return( new );
	}
	else {
	
		// Place the pixels back to the object's own list and delete
		// the flags from mark
		
		while ( ( p = DlinkRemoveFirst( &sepa ) ) ) {
		
			info->mark[ p->y ][ p->x ] = obj->id;
			DlinkAddTail( &obj->pixels, p );
		}
		
		// No new object is created
		
		return( NULL );
	}
}

/*
*	Fills holes in an object
*	~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	FixupFillHoles( INFO *info, OBJECT obj )
{
DLINK	temp, hole;	
PIXEL	p, np;
int		y, x, dy, dx, ty, tx;
int		flag;

	// Calculate the bounding box for the object
	
	obj->bbox.tly = info->ys + 1 + OFFSET;
	obj->bbox.tlx = info->xs + 1 + OFFSET;
	obj->bbox.bry = 0;
	obj->bbox.brx = 0;
	
	DlinkCat( &obj->pixels, &obj->border );
	
	for ( p = DlinkFirst( &obj->pixels ) ; p ; p = DlinkNext( p ) ) {
	
		if ( p->y < obj->bbox.tly ) obj->bbox.tly = p->y;
		if ( p->x < obj->bbox.tlx ) obj->bbox.tlx = p->x;
		if ( p->y > obj->bbox.bry ) obj->bbox.bry = p->y;
		if ( p->x > obj->bbox.brx ) obj->bbox.brx = p->x;
	}
	
	// Increase the bounding box by 2 pixels in each direction and
	// flood fill it starting at (0,0)
	
	obj->bbox.tly -= 2;
	obj->bbox.tlx -= 2;
	obj->bbox.bry += 2;
	obj->bbox.brx += 2;
	
	np = MallocPixel();
	np->y = obj->bbox.tly;
	np->x = obj->bbox.tlx;
	info->mark[ np->y ][ np->x ] |= MARK_FLAG;
	
	DlinkInit( &temp );
	DlinkAddHead( &temp, np );
	
	while ( ( p = DlinkRemoveFirst( &temp ) ) ) {
	
		for ( dy = -1 ; dy <= 1 ; dy++ ) {
		
			if ( p->y + dy < obj->bbox.tly ) continue;
			if ( p->y + dy > obj->bbox.bry ) continue;
			
			for ( dx = -1 ; dx <= 1 ; dx++ ) {
			
				if ( p->x + dx < obj->bbox.tlx ) continue;
				if ( p->x + dx > obj->bbox.brx ) continue;
				
				if ( info->mark[ p->y + dy ][ p->x + dx ] & MARK_FLAG )
				
					continue;
		
				if ( info->mark[ p->y + dy ][ p->x + dx ] == obj->id )
					
					continue;
					
				info->mark[ p->y + dy ][ p->x + dx ] |= MARK_FLAG;
				
				np = MallocPixel();
				np->y = p->y + dy;
				np->x = p->x + dx;
				
				DlinkAddTail( &temp, np );
			}
		}
		
		Free( p );
	}
	
	// Now within the bounding box anything not marked and not being
	// part of the object is a hole. However, it is possible that there
	// is a full foreign object within the hole... 
	
	for ( y = obj->bbox.tly ; y <= obj->bbox.bry ; y++ ) {
	
		for ( x = obj->bbox.tlx ; x <= obj->bbox.brx ; x++ ) {
	
			if ( ! info->mark[ y ][ x ] ) {
			
				// We found a completely empty pixel, it is hole!
				// Do an other flood fill but remember the pixels
				
				hole  = DLINK_EMPTY;
				flag  = FALSE;
				p	  = MallocPixel();
				p->x  = x;
				p->y  = y;
				
				DlinkAddTail( &temp, p  );
				
				while ( ( p = DlinkRemoveFirst( &temp ) ) ) {
				
					// See if the pixel is empty and unmarked. 
					// If not, forget it.
					
					if ( info->mark[ p->y ][ p->x ] ) {
					
						Free( p );
						continue;
					}
					
					// OK, this is a new pixel so we add it to the
					// hole list and mark it as taken.
					// DO NOT Free( p ) later!!!
					
					DlinkAddTail( &hole, p );
					info->mark[ p->y ][ p->x ] |= MARK_FLAG;
					
					// Now check the neighbours
					
					for ( dy = -1 ; dy <= 1 ; dy++ ) {
					
						for ( dx = -1 ; dx <= 1 ; dx++ ) {
						
							ty = p->y + dy;
							tx = p->x + dx;
							
							if ( info->mark[ ty ][ tx ] & MARK_FLAG ) {
							
								// This pixel is already marked, so
								// we don't consider it any further
								
								continue;
							}
							
							if ( ! info->mark[ ty ][ tx ] ) {
							
								// Empty, queue it for processing
								
								np	  = MallocPixel();
								np->y = p->y + dy;
								np->x = p->x + dx;
								DlinkAddTail( &temp, np );
								continue;
							}
							
							if ( info->mark[ ty ][ tx ] != obj->id ) {
							
								// We found a foreign object in the hole,
								// just make note of it
								
								flag = TRUE;
							}
							
							// In any other case, just forget this location
						}
					}
				}
				
				// We finished with the flood fill of the hole. The queue
				// hole contains the pixels in the hole. The flag is TRUE
				// if the hole contains a foreign object and FALSE if it 
				// had nothing just empty space. 
				
				if ( ! flag ) {
				
					// It was empty, add the pixels to the object and
					// mark them as taken by the object
					
					while ( ( p = DlinkRemoveFirst( &hole ) ) ) {
					
						info->mark[ p->y ][ p->x ] = obj->id;
						DlinkAddTail( &obj->pixels, p );
					}
				}
				else {
				
					// The hole should not be filled, just free all pixels
					// on the queue. We leave the mark flagged, so that any
					// subsequent hole search for this object would not 
					// stumble upon this hole again.
					
					while ( ( p = DlinkRemoveFirst( &hole ) ) ) {
					
						Free( p );
					}
				}
			}
		}
	}
	
	// Done. Now run through the bounding box again and remove all
	// markers
	
	for ( y = obj->bbox.tly ; y <= obj->bbox.bry ; y++ ) {
	
		for ( x = obj->bbox.tlx ; x <= obj->bbox.brx ; x++ ) {
	
			info->mark[ y ][ x ] &= ~MARK_FLAG;
		}
	}
}

/*
*	Finalises an object
*	~~~~~~~~~~~~~~~~~~~
*	Creates the bounding box, calculates the final border and
*	size. The border pixels will be on the border list AS WELL 
*	AS on the pixels list!
*/

static	void	FixupFinalise( INFO *info, OBJECT obj )
{
PIXEL	p, n;
int		x, y, d;

	FixupFindPerimeter( info, obj );
	
	obj->bbox.tly = info->ys + 1;
	obj->bbox.tlx = info->xs + 1;
	obj->bbox.bry = 0;
	obj->bbox.brx = 0;
	
	// Calculate the bounding box
		
	for ( p = DlinkFirst( &obj->border ) ; p ; p = DlinkNext( p ) ) {
	
		p->y -= OFFSET;
		p->x -= OFFSET;
	
		if ( p->y < obj->bbox.tly ) obj->bbox.tly = p->y;
		if ( p->x < obj->bbox.tlx ) obj->bbox.tlx = p->x;
		if ( p->y > obj->bbox.bry ) obj->bbox.bry = p->y;
		if ( p->x > obj->bbox.brx ) obj->bbox.brx = p->x;
	}
	
	// Offset the border pixels
	
	for ( p = DlinkFirst( &obj->border ) ; p ; p = DlinkNext( p ) ) {
	
		p->y -= obj->bbox.tly;
		p->x -= obj->bbox.tlx;
	}
	
	// Offset the area pixels
	
	for ( p = DlinkFirst( &obj->pixels ) ; p ; p = DlinkNext( p ) ) {
	
		p->y -= OFFSET + obj->bbox.tly;
		p->x -= OFFSET + obj->bbox.tlx;
	}
	
	// Add a copy of the border pixels to the area
	
	for ( p = DlinkFirst( &obj->border ) ; p ; p = DlinkNext( p ) ) {
	
		n	 = MallocPixel();
		n->y = p->y;
		n->x = p->x;
		DlinkAddTail( &obj->pixels, n );
	}
	
	// Calculate the distance division trees
	
	obj->atree = DtreeCreate( &obj->bbox, &obj->pixels );
	obj->ptree = DtreeCreate( &obj->bbox, &obj->border );
	
	// Calculate the weighted centre and the equivalent diameter
	
	x = y = d = 0;
	
	for ( p = DlinkFirst( &obj->pixels ) ; p ; p = DlinkNext( p ) ) {
		
		x += p->x * p->v;
		y += p->y * p->v;
		d += p->v;
	}
	
	obj->xc = x / (double) d;
	obj->yc = y / (double) d;
	obj->ed = sqrt( d / 255.0 );
}

/*
*	Tests if an object is on the image boundary
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Returns TRUE if it is, FALSE otherwise
*
*	An object is a boundary object if at least BORD_THOLD percent of its
*	perimeter pixels on any side are at on the image boundary.
*/

static	int		FixupIsBoundary( INFO *info, OBJECT obj )
{
int		t, l, r, b, n;
PIXEL	p;

	n = t = l = r = b = 0;
	
	FixupFindPerimeter( info, obj );
	
	for ( p = DlinkFirst( &obj->border ) ; p ; p = DlinkNext( p ), n++ ) {
		
		if ( p->x <= OFFSET + 1 ) l++;
		if ( p->y <= OFFSET + 1 ) t++;
		if ( p->x >= info->xs + OFFSET - 3 ) r++;
		if ( p->y >= info->ys + OFFSET - 3 ) b++;
	}
	
	if ( l * 100 >= BORD_THOLD * n || r * 100 >= BORD_THOLD * n || 
		 t * 100 >= BORD_THOLD * n || b * 100 >= BORD_THOLD * n )  
		
		return( TRUE );
	else
		return( FALSE );
}

