/****************************************************************************/
/* Copyright 2004, © Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Object processing														*/
/*	~~~~~~~~~~~~~~~~~														*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltán Kócsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#ifndef __PROCESS_INTERN_H__
#define	__PROCESS_INTERN_H__

#include <types.h>
#include <objects.h>
#include <distance.h>

/****************************************************************************/
/*							Internal definitions							*/
/****************************************************************************/

#define	MAXLEV			12		/* Maximum height of the tree (8Kx8K image)	*/

#define	DIST_INF		0x7fffffff

/****************************************************************************/
/*							Module local types								*/
/****************************************************************************/

/*
*	A pixel belonging to a feature
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

struct s_pixel {

	DLINK	link;				/* Link pointers							*/
	short	y, x;				/* Coordinates of the pixel					*/
	ubyte	v;					/* Value of the pixel						*/
};

/*
*	Spatial division tree node
*	~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

typedef struct s_sdiv {

	short			s;			/* Size of a quadrant						*/
	short			x, y;		/* Top left corner coordinates				*/
	short			m;			/* Bit map of active quadrants				*/
	struct s_sdiv	*quad[ 4 ];	/* Pointers to the 4 quadrants				*/

} SDIV;

/*
*	Feature on the image
*	~~~~~~~~~~~~~~~~~~~~
*/


struct s_object {

	DLINK	link;				/* Link pointers							*/
	DLINK	pixels;				/* List of pixels belonging to the feature	*/
	DLINK	border;				/* List of pixels bounding the feature		*/
	RECT	bbox;				/* Bounding box of the feature				*/
	int		area;				/* Number of pixels in the feature			*/
	int		peri;				/* Number of pixels in perimeter			*/
	int		id;					/* Unique identifier						*/
	SDIV	*atree;				/* Distance division tree root, all			*/
	SDIV	*ptree;				/* Distance division tree root, perimeter	*/
	double	xc, yc;				/* Weighted centre of object				*/
	double	ed;					/* Equivalent diameter						*/
	
	short	min, max;			/* Transient, pixel acceptance values		*/
	PIXEL	curr;				/* Transient, pixel currently processed		*/
};

/*
*	Triangle
*	~~~~~~~~
*/

struct s_tria {

	DLINK	link;				/* Link pointers							*/
	OBJECT	beg;				/* Beginning of line segment				*/
	OBJECT	mid;				/* Mid point of line segment, if exists		*/
	OBJECT	end;				/* End of line segment						*/

};

/****************************************************************************/
/*							Inline functions								*/
/****************************************************************************/

/*
*	Allocates a object descriptor
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	inline	OBJECT	MallocObject( void )
{
	return( (OBJECT) MallocPool( memObject, sizeof( struct s_object ) ) );
}

/*
*	Allocates a pixel descriptor
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	inline	PIXEL	MallocPixel( void )
{
	return( (PIXEL) MallocPool( memPixel, sizeof( struct s_pixel ) ) );
}

/*
*	Allocates a divider node
*	~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	inline	SDIV	*MallocSdiv( void )
{
	return( (SDIV *) MallocPool( memSdiv, sizeof( struct s_sdiv ) ) );
}

/*
*	Allocates a distance pair node
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	inline	DIST	*MallocDistance( void )
{
	return( (DIST*) MallocPool( memDist, sizeof( struct s_dist ) ) );
}

/*
*	Allocates an overlap area descriptor
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	inline	OVLP	*MallocOverlap( void )
{
	return( (OVLP*) MallocPool( memOvlp, sizeof( struct s_ovlp ) ) );
}

/*
*	Allocates a triangle descriptor
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	inline	TRIA	MallocTria( void )
{
	return( (TRIA) MallocPool( memTria, sizeof( struct s_tria ) ) );
}

/****************************************************************************/
/*					Interface functions within the module					*/
/****************************************************************************/

/*
*	Deletes a distance division tree
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	DtreeDestroy( SDIV *tree );

/*
*	Builds a distance division tree
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

SDIV	*DtreeCreate( RECT *bbox, DLINK *pixels );

/*
*	Finds the minimum distance between two objects
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	If the two objects are more than limit apart, the returned value
*	is DIST_INF. Otherwise, the returned value is the minimum distance
*	and the RECT structure will contain the matching pixels.
*/

int		DistMin( OBJECT ref, OBJECT act, RECT *coords, int maxd );

/*
*	Finds the overlap between objects
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Finds the overlap between two objects. Returns the number of
*	pixels overlapping. If the 'pixs' pointer is not NULL, then
*	a list of pixels will be placed onto the list, indicating
*	the overlap positions.
*/

int		DistOvl( OBJECT ref, OBJECT act, DLINK *pixs );

#endif

