/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Object extraction - fixup procedures									*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~									*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include <types.h>
#include <memory.h>
#include <dlinklist.h>
#include <objects.h>
#include <process.h>
#include <progress.h>
#include <distance.h>
#include <randomlist.h>
#include "process_intern.h"

/****************************************************************************/
/*						Local definitions									*/
/****************************************************************************/

#define	MAXATTEMPTS		20
#define	MAXTRIES		100

/****************************************************************************/
/*						Prototypes											*/
/****************************************************************************/

static	void	RandomCoords( RECT *box, short *y, short *x );

/****************************************************************************/
/*						Local definitions									*/
/****************************************************************************/

/*
*	Creates a random object series based on the template
*
*	Parameters:
*
*		box		The bounding box of the useful area
*		res		The resulting list of objects
*		tmpl	Template list
*		conf	Object list of the confinement area
*
*	Return value:
*
*		A flag indicating success of failure
*
*	Although the newly created object list looks like as if it was
*	a true object list, in fact it is just a reference list to the
*	template list. Therefore, it is imperative that it is *NOT* freed
*	by the ObjectFree() or ObjectFreeList() functions but only using
*	RandomListFree function, provided in this file.
*/

int		RandomListCreate( RECT *box, DLINK *res, DLINK *tmpl, DLINK *conf )
{
OBJECT	old, new, obj;
RECT	dummy;
int		attempt, try;
int		xs, ys;

	DlinkInit( res );
	
	for ( attempt = 0 ; attempt < MAXATTEMPTS ; attempt++ ) {
		
		for ( old = DlinkFirst( tmpl ) ; old ; old = DlinkNext( old ) ) {
	
			// Get a new descriptor
		
			new = MallocObject();
		
			// Inherit everything from the template descriptor.
			// This will basically create duplicate references to 
			// the old object's pixel and spatial trees
		
			*new = *old;
		
			// Calculate the size of the bounding box for the object
		
			ys = new->bbox.bry - new->bbox.tly;
			xs = new->bbox.brx - new->bbox.tlx;
		
			// Shrink the image bounding box accordingly
		
			box->bry -= ys;
			box->brx -= xs;
			
			for ( try = 0 ; try < MAXTRIES ; try++ ) {
		
				// Create a random x, y coordinate pair within the image 
				// bounding box and accomodate 
			
				RandomCoords( box, &new->bbox.tly, &new->bbox.tlx );
			
				new->bbox.bry = new->bbox.tly + ys;
				new->bbox.brx = new->bbox.tlx + xs;
				
				// Now check if the new object is in clash with the
				// confinement list
			
				for ( obj=DlinkFirst( conf ) ; obj ; obj=DlinkNext( obj ) ) {
			
					if ( DistMin( obj, new, &dummy, 0 ) != DIST_INF ) {
				
						// It touches a confinement object
					
						break;
					}
				}
			
				if ( obj ) {
			
					// No good, try again
				
					continue;
				}
			
				// Let's see if it clashes with any of the already placed
				// objects. Unlike the confinement restrictions, between
				// objects the minimum distance is at least 1 pixel, that 
				// is 4 in distance numbers (at least 2 pixels difference 
				// in one direction)
			
				for ( obj = DlinkFirst(res) ; obj ; obj = DlinkNext( obj ) ) {
			
					if ( DistMin( obj, new, &dummy, 4 ) != DIST_INF ) {
				
						// It is too close to an other object
					
						break;
					}
				}
			
				if ( obj ) {
			
					// Too close to an already placed one try again
				
					continue;
				}
				
				// Success! We can add this object onto our list and leave
				
				DlinkAddTail( res, new );
				break;
			}
			
			// Restore the original bounding box
			
			box->bry += ys;
			box->brx += xs;
			
			if ( try == MAXTRIES ) {
			
				// We could not add the next object to the list.
				// Release the descriptor, then the list we've
				// built so far, and possibly try it from the 
				// beginning
				
				Free( new );
				RandomListFree( res );
				break;
			}				
		}
		
		if ( ! old ) {
		
			// We succeeded
			
			return( TRUE );
		}
		
		// We failed, we start it all over again
	}
	
	// Report failure
	
	return( FALSE );
}

/*
*	Frees a random list
*	~~~~~~~~~~~~~~~~~~~
*	Parameters:
*
*		list	The list to free
*
*	Return value:
*		-
*	
*/

void	RandomListFree( DLINK *list )
{
OBJECT	obj;

	while ( ( obj = DlinkRemoveFirst( list ) ) ) Free( obj );
}

/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/

static	void	RandomCoords( RECT *box, short *y, short *x )
{
	*y = random() % ( box->bry - box->tly )	+ box->tly;
	*x = random() % ( box->brx - box->tlx )	+ box->tlx;
}
