/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Object extraction														*/
/*	~~~~~~~~~~~~~~~~~~														*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>

#include <types.h>
#include <dlinklist.h>
#include <objects.h>
#include <memory.h>
#include <image.h>
#include <options.h>
#include "process_intern.h"
#include "distfunc.h"

/****************************************************************************/
/*							Local types										*/
/****************************************************************************/

typedef	struct {

	RECT	bbox;
	int		ys, xs;
	PIXEL	**mark;
	
} INFO;

/****************************************************************************/
/*							Prototypes										*/
/****************************************************************************/

static	SDIV	*DtreeBuild( INFO *info, int lev, int y, int x );
static	void	DtreeDelete( SDIV *node, int lev );

/****************************************************************************/
/*						Interface functions									*/
/****************************************************************************/

/*
*	Deletes a distance division tree
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	DtreeDestroy( SDIV *tree )
{
	if ( tree ) DtreeDelete( tree, MAXLEV );
}

/*
*	Builds a distance division tree
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

SDIV	*DtreeCreate( RECT *bbox, DLINK *pixels )
{
INFO	info;
PIXEL	p;
SDIV	*tree;

	info.bbox.tly = 0;
	info.bbox.tlx = 0;
	info.bbox.bry = bbox->bry - bbox->tly;
	info.bbox.brx = bbox->brx - bbox->tlx;
	info.ys		  = info.bbox.bry + 1;
	info.xs		  = info.bbox.brx + 1;
	info.mark	  = MallocArray( info.ys, info.xs, sizeof( void * ) );
	
	for ( p = DlinkFirst( pixels ) ; p ; p = DlinkNext( p ) ) {
		
		info.mark[ p->y ][ p->x ] = p;
	}
	
	tree = DtreeBuild( &info, MAXLEV, 0, 0 );
	Free( info.mark );
	return( tree );
}

/****************************************************************************/
/*							Local functions									*/
/****************************************************************************/

/*
*	Recursively builds the spatial division tree for a object
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	SDIV	*DtreeBuild( INFO *info, int lev, int y, int x )
{
RECT	me;
SDIV	*n;
int		xt, yt, i, size;
static	int	yoff[] = { 0, 0, 1, 1 };
static	int	xoff[] = { 0, 1, 0, 1 };

	if ( ! lev ) {
	
		// We are down at pixel level
		
		n	 = MallocSdiv();
		n->x = x;
		n->y = y;
		n->s = 1;
		n->m = 0;
		
		for ( i = 0 ; i < 4 ; i++ ) {
		
			yt = y + yoff[ i ];
			xt = x + xoff[ i ];
			
			// We do not have to test for yt,xt < 0	condition because
			// the algorithm guarantess that that can never happen
			
			if ( yt < info->ys && xt < info->xs && info->mark[ yt ][ xt ] ) {
		
				// We found a pixel
				
				n->quad[ i ] = (SDIV *) info->mark[ yt ][ xt ];
				n->m		|= 1 << i;
			}
			else {
			
				n->quad[ i ] = NULL;
			}
		}
		
		if ( ! n->m ) {
		
			Free( n );
			n = NULL;
		}
	}
	else {
	
		// Don't worry about pixels, just check overlaps
		
		size   = ( 1 << ( lev + 1 ) ) - 1;
		me.tly = y;
		me.tlx = x;
		me.bry = y + size;
		me.brx = x + size;
		
		if ( ! RectOverlap( &info->bbox, &me ) ) {
	
			// No overlap between the current square and the object
		
			n = NULL;
		}
		else {
	
			// There's some overlap, divide the space and go down
					
			size = 1 << lev;
			n	 = MallocSdiv();
			n->y = y;
			n->x = x;
			n->s = size;
			n->m = 0;
			
			n->quad[ 0 ] = DtreeBuild( info, lev-1, y,		x	   );
			n->quad[ 1 ] = DtreeBuild( info, lev-1, y,		x+size );
			n->quad[ 2 ] = DtreeBuild( info, lev-1, y+size, x	   );
			n->quad[ 3 ] = DtreeBuild( info, lev-1, y+size, x+size );
			
			// Keep the box only if it has something in it
			
			if ( n->quad[ 0 ] ) n->m |= 1;
			if ( n->quad[ 1 ] ) n->m |= 2;
			if ( n->quad[ 2 ] ) n->m |= 4;
			if ( n->quad[ 3 ] ) n->m |= 8;
		
			if ( ! n->m ) {
			
				Free( n );
				n = NULL;
			}
		}
	}
	
	return( n );
}

/*
*	Recursively deletes a division tree
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	DtreeDelete( SDIV *node, int lev )
{
int		i;

	if ( ! lev ) {
	
		if ( node ) Free( node );
	}
	else {
	
		for ( i = 0 ; i < 4 ; i++ ) {
		
			if ( node->quad[ i ] ) {
			
				DtreeDelete( node->quad[ i ], lev - 1 );
			}
		}
		
		Free( node );
	}
}

#if 0
/*
*	Clears the marker array
*	~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	DtreeClearMark( INFO *info, int ymax, int xmax )
{
int		y, x;

	for ( y = 0 ; y < ymax ; y++ ) {
	
		for ( x = 0 ; x < xmax ; x++ ) {
		
			info->mark[ y ][ x ] = 0;
		}
	}
}

#endif
