/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	GUI actions box creation												*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~												*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <types.h>
#include <proxan.h>
#include <dlinklist.h>
#include <distance.h>
#include <memory.h>
#include <stats.h>

#include "process_intern.h"


/****************************************************************************/
/*						Local definitions									*/
/****************************************************************************/

#define	MIN_BIN_SIZE		10
#define	I_STEPS				500

/****************************************************************************/
/*						Prototypes											*/
/****************************************************************************/

static	void	StatFrequency( SINFO info );
static	void	StatChisquare( SINFO info );
static	void	StatUsrcurves( SINFO info, STAT *res );

static	double	PoissonDist( int x, double mu );
static	double	ChiSquareDens( int n, double x );
static	double	ChiSquareDist( int n, double x );

/****************************************************************************/
/*						Prototypes											*/
/****************************************************************************/

struct s_sinfo {
	
	int		rows;
	int		cols;
	int		maxd;
	int		maxo;
	int		bins;
	double	*chis;
	int		**freq;
	DLINK	lists;
};
	
typedef	struct {

	DLINK	link;
	DLINK	dists;
	DLINK	ovlps;
	int		min, max;

} SAMPLE;

/*
*	Prepares the statistical analysis
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters:
*		dist		A doubly-linked list of DIST objects that describe
*					the distances of the sample data
*		ovlp		A doubly-linked list of OVLP objects that describe
*					the overlaps of the sample data
*		maxd		The maximum distance on the list
*		tests		The number of tests that will be performed
*
*	Return value:
*		A statistical data structure pointer
*/

SINFO	StatPrepare( DLINK *dist, DLINK *ovlp )
{
SINFO	info;
SAMPLE	*first;
OVLP	*op, *ot;
DIST	*dp, *dt;

	info = Malloc( sizeof( struct s_sinfo ) );
	DlinkInit( &info->lists );
	info->rows = 0;
	info->maxd = 0;
	info->maxo = 0;
	
	first = Malloc( sizeof( SAMPLE ) );
	DlinkInit( &first->dists );
	DlinkInit( &first->ovlps );
	
	for ( dp = DlinkFirst( dist ) ; dp ; dp = DlinkNext( dp ) ) {
	
		dt = MallocDistance();
		*dt = *dp;
		DlinkAddTail( &first->dists, dt );
		if ( info->maxd < dp->dist ) info->maxd = dp->dist;
	}
	
	for ( op = DlinkFirst( ovlp ) ; op ; op = DlinkNext( op ) ) {
	
		ot = MallocOverlap();
		*ot = *op;
		DlinkAddTail( &first->ovlps, ot );
		if ( info->maxo < op->nump ) info->maxo = op->nump;
	}
	
	DlinkAddTail( &info->lists, first );
	info->rows++;
	return( info );
}

/*
*	One run of the new data
*	~~~~~~~~~~~~~~~~~~~~~~~
*	Adds a new set of random data to the statistical pool
*	
*	Parameters:
*		sdesc		The statistical data obtained by StatPrepare()
*		dist		A doubly linked list of the random distances
*		ovlp		A doubly-linked list of the random overlaps
*	Return value:
*		-
*
*	The dist and ovlp lists are retained by the routine and the relevant
*	DLINK objects will be empty.
*/

void	StatRound( SINFO info, DLINK *dist, DLINK *ovlp )
{
OVLP	*op;
DIST	*dp;
SAMPLE	*next;

	for ( dp = DlinkFirst( dist ) ; dp ; dp = DlinkNext( dp ) ) {
	
		if ( info->maxd < dp->dist ) info->maxd = dp->dist;
	}
	
	for ( op = DlinkFirst( ovlp ) ; op ; op = DlinkNext( op ) ) {
	
		if ( info->maxo < op->nump ) info->maxo = op->nump;
	}
	
	next = Malloc( sizeof( SAMPLE ) );
	DlinkInit( &next->dists );
	DlinkInit( &next->ovlps );
	DlinkCat( &next->dists, dist );
	DlinkCat( &next->ovlps, ovlp );
	DlinkAddTail( &info->lists, next );
	info->rows++;
}

/*
*	Releases the info structure
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters:
*		info		The statistical data obtained by StatPrepare()
*
*	Return value:
*		-
*/

void	StatDestroy( SINFO info )
{
SAMPLE	*sp;
void	*tp;

	while ( ( sp = DlinkRemoveFirst( &info->lists ) ) ) {
	
		while ( ( tp = DlinkRemoveFirst( &sp->dists ) ) ) Free( tp );
		while ( ( tp = DlinkRemoveFirst( &sp->ovlps ) ) ) Free( tp );
		Free( sp );
	}
	
	Free( info );
}

/*
*	Calculates statistical results from the gathered data
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Parameters:
*		info		The statistical data obtained by StatPrepare()
*
*	Return value:
*		A Malloc-ed structure containing the statistical results
*		or NULL, if the statistics could not be calculated.
*
*	The dist and ovlp lists are retained by the routine and the relevant
*	DLINK objects will be empty.
*/

STAT	*StatFinal( SINFO info )
{
STAT	*res;
	
	res = Malloc( sizeof( STAT ) );
	
	StatFrequency( info );
	StatChisquare( info );
	StatUsrcurves( info, res );	

	res->min = info->maxo;
	res->max = info->maxd;
	res->pro = 1.0 - ChiSquareDist( info->bins-1, info->chis[ 0 ] );
	return( res );
}

/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/

/*
*	Converts the data into a frequency table
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	StatFrequency( SINFO info )
{
int		row, col;
SAMPLE	*sp;
OVLP	*op;
DIST	*dp;

	// Get the measurements into physical space
	
	info->maxd = sqrt( (double) info->maxd ) + 0.5;
	info->maxo = sqrt( (double) info->maxo ) + 0.5;
	
	// Distances can come between [0,maxd] and overlaps come
	// between [0,maxo]. 0 overlap is irrelevant and 0 distances
	// can be thrown away, because they represent an overlap.
	// Therefore, to form a continuous distribution, we want
	// 0 distance to map to 1 pixel overlap. Then we want the
	// whole thing to be shifted up so that the array index
	// 0 represents maxo, the array index maxo-1 represents the
	// overlap of 1, maxo the distance 1 and maxo+maxd-1 would
	// select maxd. Thus, the array must have maxo+maxd elements.
	
	info->cols = info->maxo + info->maxd;
	info->freq = MallocArray( info->rows, info->cols, sizeof( int ) );
	
	// Fill the frequency array and in the meantime we can release
	// all the allocated structures
	
	for ( row = 0 ; ( sp = DlinkRemoveFirst( &info->lists ) ) ; row++ ) {
	
		while ( ( dp = DlinkRemoveFirst( &sp->dists ) ) ) {
		
			if ( ( col = sqrt( (double) dp->dist ) + 0.5 ) ) {
			
				info->freq[ row ][ col + info->maxo - 1 ]++;
			}
			
			Free( dp );
		}
		
		while ( ( op = DlinkRemoveFirst( &sp->ovlps ) ) ) {
		
			if ( ( col = sqrt( (double) op->nump ) + 0.5 ) ) {
			
				info->freq[ row ][ info->maxo - col ]++;
			}
			
			Free( op );
		}
		
		Free( sp );
	}
}

/*
*	Calculates the chi-square value for each row
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	StatChisquare( SINFO info )
{
double	*mean;
int		col, row, n;
int		bin, beg, end;
double	sum;

	// Allocate and initialise our data sets
	
	info->chis = Malloc( info->rows * sizeof( double ) );
	mean = Malloc( info->cols * sizeof( double ) );
	
	memset( info->chis, 0, info->rows * sizeof( double ) );
	memset( mean, 0, info->cols * sizeof( double ) );
	
	// Claculate the mean frequency for each column
	
	for ( col = 0 ; col < info->cols ; col++ ) {
	
		n = 0;
		
		for ( n = 0, row = 1 ; row < info->rows ; row++ ) {
		
			n += info->freq[ row ][ col ];
		}
			
		mean[ col ] = (double) n / ( info->rows - 1 );
	}
	
	// Now we can start calculating the chi-squares for each row. 
	// We keep accumulating columns until we get at least 10 in mean
	// in each bin.
	
	for ( bin = 0, sum = 0, beg = 0, end = 0 ; end < info->cols ; end++ ) {
	
		sum += mean[ end ];
		if ( sum < MIN_BIN_SIZE ) continue;
		
		// The current bin has an overall mean of 'sum' and is composed
		// of columns beg..end inclusive. Calculate the partial sums for
		// each row. In this calculation we take the raw data as well
		
		for ( row = 0 ; row < info->rows ; row++ ) {
		
			for ( n = 0, col = beg ; col <= end ; col++ ) {
			
				n += info->freq[ row ][ col ];
			}
			
			info->chis[ row ] += ( n - sum ) * ( n - sum ) / sum;
		}
		
		// We increment the bin count and prepare for the next round
		
		bin++;
		sum = 0;
		beg = end + 1;
	}
	
	// Pass back the bin count
	
	info->bins = bin;
	
	// We can get rid of the means
	
	Free( mean );
}

/*
*	Calculates the means and the deviations for the samples
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	StatUsrcurves( SINFO info, STAT *res )
{
double	step, curr, n, min, max;
int		row, col, i, beg, pnt;
int		*nums;

	nums = Malloc( info->rows * sizeof( int ) );
	memset( nums, 0, info->rows * sizeof( int ) );
	
	for ( row = 0 ; row < info->rows ; row++ ) {
	
		for ( col = 0 ; col < info->cols ; col++ ) {
		
			nums[ row ] += info->freq[ row ][ col ];
		}
	}
	
	for ( i = 0 ; i < STAT_POINTS ; i++ ) {
	
		res->pnts[ MEAN ][ i ] = 0;
		res->pnts[ SMIN ][ i ] = 0;
		res->pnts[ SMAX ][ i ] = 0;
		res->pnts[ ORIG ][ i ] = 0;
	}
	
	step = (double) ( info->cols ) / (STAT_POINTS - 1);
	curr = step / 2;
	pnt  = 0;
	
	for ( beg = col = 0 ; col < info->cols ; col++ ) {
	
		if ( col < info->cols - 1 && col < curr ) continue;
		
		min = 1E6;
		max = 0;
		
		for ( row = 1 ; row < info->rows ; row++ ) {
		
			for ( n = 0.0, i = beg ; i <= col ; i++ ) {
			
				n += (double) info->freq[ row ][ i ] / nums[ row ];
			}
			
			if ( n > max ) max = n;
			if ( n < min ) min = n;
			
			res->pnts[ MEAN ][ pnt ] += n / ( info->rows - 1 );
		}
		
		res->pnts[ SMIN ][ pnt ] = min;
		res->pnts[ SMAX ][ pnt ] = max;
		
		for ( n = 0.0, i = beg ; i <= col ; i++ ) {
			
			res->pnts[ ORIG ][ pnt ] += (double) info->freq[0][i] / nums[0];
		}
		
		pnt++;
		beg   = col + 1;
		curr += step;
	}
	
	Free( nums );
}

/****************************************************************************/
/*						Statistical functions								*/
/****************************************************************************/

/*
*	Calculates the chi-square's density function
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	The parameter n is the degree of freedom.
*
*
*            (n/2-1)    -n/2
*           x        * e
*	chi^2=  -----------------------
*            n/2
*           2        * GAMMA( n/2 )
*
*	Note that the gamma(x) function in the C library returns
*	ln( GAMMA(x) ) not GAMMA(x).
*/

static	double	ChiSquareDens( int n, double x )
{
double	n2;

	if ( x <= 0 ) return( 0.0 );
	
	n2 = n / 2.0;
	
	return( exp( ( log( x ) * ( n2 - 1 ) - ( x / 2 ) ) -  
		       ( log( 2 ) *  n2 + gamma( n2 ) ) ) );
}

/*
*	Calculates the chi-square's probability distribution function
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	This function is the definite integral of the density function
*	between 0 and the given value.
*	The parameter n is the degree of freedom.
*
*	We use the Simpson formula to integrate the density function.
*/

static	double	ChiSquareDist( int n, double x )
{
int		i;
double  xn, yn;
double  stp, sum;

	if ( x > 50 ) return( 1 );
	
	stp = x / ( I_STEPS - 1 );
        
	for ( sum = 0, i= 1 ; i < I_STEPS - 1 ; i++ ) {
        
		xn	 = stp * i;
		yn	 = ChiSquareDens( n, xn );
		sum += yn + ( ( i & 1 ) ? 0 : yn );
	}
	
	sum += sum;
	sum += ChiSquareDens( n, 0 );
	sum += ChiSquareDens( n, x );
        
	return( sum * stp / 3 );
}

/*
*	Calculates the Poisson distribution 
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*				  -mu     x
*				e     * mu       ( -mu + x * ln(mu) - ln(x!) )
*	p(x;mu) = -------------- = e
*					x!
*
*	We also use the fact that GAMMA( x ) = (x-1)!
*	Note that the 'gamma()' library function calculates ln(GAMMA()),
*	despite its name.
*/

static	double	PoissonDist( int x, double mu )
{
	if ( x < 0 ) return( 0 );
	if ( ! mu  ) return( x ? 0 : 1 );
	return( exp( -mu + x * log( mu ) - gamma( (double) (x + 1) ) ) );
}
