/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Basic image related operation											*/
/*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~											*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <tiffio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "types.h"
#include "memory.h"
#include "image.h"

/****************************************************************************/
/*						Prototypes											*/
/****************************************************************************/

static	IMAGE	*TiffRead( const char *name );

/****************************************************************************/
/*						Public functions									*/
/****************************************************************************/

/*
*	Creates an image of the given size
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

IMAGE	*ImageCreate( int y, int x, int num_layer )
{
IMAGE	*im;
int		i, j;

	im = Malloc( sizeof( IMAGE ) );
	
	im->x = x;
	im->y = y;
	im->l = num_layer;
		
	for ( i = 0 ; i < MAX_LAYER ; i++ ) {
	
		if ( i < num_layer )
			
			im->im[ i ] = MallocArray( y, x, sizeof( ubyte ) );
		else
			im->im[ i ] = NULL;
			
	}
	
	return( im );
}

/*
*	Clears a layer of an image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	ImageClear( IMAGE *im, int layer, ubyte val )
{
int		x, y;

	for ( y = 0 ; y < im->y ; y++ ) {
	
		for ( x = 0 ; x < im->x ; x++ ) {
	
			im->im[ layer ][ y ][ x ] = val;
		}
	}
}

/*
*	Copies a layer to an image
*	~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	ImageCopyLayer( IMAGE *new, int newl, IMAGE *old, int oldl )
{
int		y;

	for ( y = 0 ; y < old->y ; y++ ) {
	
		memcpy( new->im[ newl ][ y ], old->im[ oldl ][ y ], old->x );
	}
}

/*
*	Destroys an image
*	~~~~~~~~~~~~~~~~~
*/

void	ImageDestroy( IMAGE *image )
{
int		i;

	for ( i = 0 ; i < MAX_LAYER ; i++ ) {
	
		if ( image->im[ i ] ) Free( image->im[ i ] );
	}
	
	Free( image );	
}

/*
*	Duplicates an image
*	~~~~~~~~~~~~~~~~~~~
*/

IMAGE	*ImageDup( IMAGE *src )
{
IMAGE	*dst;
int		y;
int		l;

	dst = ImageCreate( src->y, src->x, src->l );
	
	for ( l = 0 ; l < src->l ; l++ ) {
	
		for ( y = 0 ; y < src->y ; y++ ) {
			
			memcpy( dst->im[ l ][ y ], src->im[ l ][ y ], src->x );
		}
	}

	return( dst );
}

/*
*	Loads an image from disk
*	~~~~~~~~~~~~~~~~~~~~~~~~
*/

IMAGE	*ImageLoad( const char *name )
{
struct stat fst;
IMAGE		*p;

	stat( name, &fst );
	if ( ! S_ISREG( fst.st_mode ) ) return( NULL );	
	p = TiffRead( name );
	return( p );
}

/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/

/*
*	Reads a TIFF image from disk
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Returns an image appropriately sized and filled.
*/

static	IMAGE	*TiffRead( const char *name )
{
TIFF			*handle;
TIFFRGBAImage	tiffim;
char			emsg[ 1024 ];
int				*raster;
IMAGE			*image;
int				x, y, i;

	TIFFSetErrorHandler( NULL );
	
	if ( ! ( handle = TIFFOpen( name, "r" ) ) ) {
		
		return( NULL );
	}
	
	if ( ! TIFFRGBAImageOK( handle, emsg ) ) {
	
		TIFFClose( handle );
		return( NULL );
	}
	
	if ( ! TIFFRGBAImageBegin( &tiffim, handle, 1, emsg ) ) {
	
		TIFFClose( handle );
		return( NULL );
	}
	
	raster = Malloc( tiffim.width * tiffim.height * 4 );
	
	if ( ! TIFFRGBAImageGet( &tiffim, (uint32 *) raster, 
							 tiffim.width, tiffim.height ) ) {
	
		Free( raster );
		TIFFClose( handle );
		return( NULL );
	}
	
	image = ImageCreate( tiffim.height, tiffim.width, 3 );
	
	for ( y = image->y - 1, i = 0 ; y >= 0 ; y-- ) {
	
		for ( x = 0 ; x < image->x ; x++, i++ ) {
		
			image->im[ R ][ y ][ x ] = TIFFGetR( raster[ i ] );
			image->im[ G ][ y ][ x ] = TIFFGetG( raster[ i ] );
			image->im[ B ][ y ][ x ] = TIFFGetB( raster[ i ] );
		}
	}

	TIFFRGBAImageEnd( &tiffim );
	TIFFClose( handle );
	Free( raster );
	return( image );
}

/*
*	Writes an RGB buffer to the disk
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*	Returns a success code
*/

int		TiffWrite( ubyte *buffer, int y, int x, const char *name )
{
TIFF	*handle;
int		i;

 	TIFFSetErrorHandler( NULL );
	
	if ( ! ( handle = TIFFOpen( name, "w" ) ) ) return( FALSE );
	
	TIFFSetField( handle, TIFFTAG_BITSPERSAMPLE, 8 );
	TIFFSetField( handle, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT );
	TIFFSetField( handle, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG );
	TIFFSetField( handle, TIFFTAG_COMPRESSION, COMPRESSION_NONE );
	TIFFSetField( handle, TIFFTAG_IMAGEWIDTH, x );
	TIFFSetField( handle, TIFFTAG_IMAGELENGTH, y );
	TIFFSetField( handle, TIFFTAG_SAMPLESPERPIXEL, 3 );
	TIFFSetField( handle, TIFFTAG_ROWSPERSTRIP, y );
	TIFFSetField( handle, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB );
	TIFFSetField( handle, TIFFTAG_SOFTWARE, "Proxan by Bendor Research Pty. Ltd." );

	for ( i = 0 ; i < y ; i++ ) {	
	
		if ( TIFFWriteScanline( handle, buffer, i, 0 ) == -1 ) {
		
			return( FALSE );
		}
		
		buffer += 3 * x;
	}
	
	TIFFClose( handle );
	return( TRUE );
}
