/****************************************************************************/
/* Copyright 2004,  Bendor Research Pty. Ltd.			All Rights Reserved	*/
/****************************************************************************/
/*																			*/
/*	Memory management														*/
/*	~~~~~~~~~~~~~~~~~														*/
/*																			*/
/*	Who				When			Comment									*/
/*																			*/
/*	Zoltn Kcsi	2004 Mar 18		Original code							*/
/*																			*/
/****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "types.h"
#include "memory.h"

/****************************************************************************/
/*							Local types										*/
/****************************************************************************/

/*
*	We always allocate more than the requested amount. For objects that
*	are small and alloced and freed often enough, we keep a pool and
*	the objects never get really freed, just put back onto the pool.
*	In addition, we do one more trick that can help us to release
*	2D arrays. When we Free() an object, if its list pointer is not
*	NULL, then it belongs to the pool pointed by the list. In that case
*	the next pointer MUST be NULL. If the list pointer is NULL, and
*	the next pointer is NULL as well, then the object is just a block
*	of memory and can be given back to the system. If, however, the
*	next pointer is non-NULL, then the object is an array of pointers
*	to Malloc() -ed blocks of memory and the next pointer casted to an
*	ulong tells as how many pointers in the list.
*/

typedef	struct	s_mem {

	struct s_mem	*next;				/* Next element on the free list	*/
	struct s_mem	**list;				/* Pool that the object belongs		*/
	int				size;				/* Size of the allocated memory		*/
	int				data[];				/* The data that the user sees		*/
	
} NODE;

typedef	struct	s_tra {

	int		size;
	int		outs;
	
} TRACK;

#define	MAXTRA	30

static	TRACK	track[ MAXTRA ] = {{0,0}};

static	void	TrackAlloc( int size )
{
int i;

	for ( i = 1 ; i < MAXTRA ; i++ ) {
	
		if ( track[ i ].size == size ) {
		
			track[ i ].outs++;
			return;
		}
	}
	
	for ( i = 1 ; i < MAXTRA ; i++ ) {
	
		if ( ! track[ i ].size ) {
		
			track[ i ].size = size;
			track[ i ].outs = 1;
			return;
		}
	}
	
	track[ 0 ].outs += size;
}

static	void	TrackFree( int size )
{
int i;

	for ( i = 1 ; i < MAXTRA ; i++ ) {
	
		if ( track[ i ].size == size ) {
			
			if ( ! --track[ i ].outs ) track[ i ].size = 0;
			return;
		}
	}
	
	track[ 0 ].outs -= size;
}

void	MallocTrackDump( char *s )
{
int 	i;
int 	all;

	printf("\n\nMEMORY DUMP: %s\n\n", s );
	printf( "Unknown size outstanding : %10d\n", track[ 0 ].outs );
	
	for ( i = 1 ; i < MAXTRA ; i++ ) {
	
		if ( ! track[ i ].size ) continue;
			

		printf( "Size=%8d out=%8d total=%10d\n",
				track[ i ].size, track[ i ].outs, 
				track[ i ].size * track[ i ].outs );
	}
}

/****************************************************************************/
/*							Prototypes										*/
/****************************************************************************/

static	void	*AllocPool( NODE **pool, int objsize );

/****************************************************************************/
/*							Local data										*/
/****************************************************************************/

NODE	*pools[ memLast ] = { NULL };	/* Object pool pointers				*/

/****************************************************************************/
/*							Public functions								*/
/****************************************************************************/

/*
*	Allocates an element from a pool
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	*MallocPool( memType pool, int size )
{
	return( AllocPool( pools + pool, size ) );
}

/*
*	Allocates an array of objects
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	*MallocArray( int y, int x, int size )
{
void	**a;

	(-1)[ (NODE *) ( a = Malloc( y * sizeof( void * ) ) ) ].next = (void *) y;
	x *= size;
	while ( y-- ) a[ y ] = memset( Malloc( x ), 0, x );
	return( a );
}

/*
*	Allocates a block of memory
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	*Malloc( int size )
{
NODE	*node;

	if ( ! ( node = malloc( sizeof( NODE ) + size ) ) ) {
	
		// We ran out of memory, fatal error
		
	}
	
	node->list = NULL;
	node->next = NULL;
	node->size = size + sizeof( NODE );
TrackAlloc( node->size );
	return( (void *) node->data );
}

/*
*	Deallocates a bit of memory
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void	Free( void *mem )
{
NODE	*ptr;
void	**lst;
int		n;

	ptr = (NODE *) mem - 1;
TrackFree( ptr->size );
	
	if ( ptr->list ) {
	
		// Object belongs to a pool
		
		if ( ptr->next ) {
	
			// Internal error!!! Object is already on a list
printf( "DEAD\n" );
		}
		
		ptr->next  = *ptr->list;
		*ptr->list = ptr;
	
	}
	else {
	
		if ( ptr->next ) {
		
			// Object is an array of Malloc()-ed blocks.
			// Recursively free all objects on the list
			
			for ( n = (int) ptr->next, lst = (void *) ptr->data ; n-- ; ) {
			
				Free( *lst++ );
			}
		}
		
		free( ptr );
	}
}

/****************************************************************************/
/*							Local functions									*/
/****************************************************************************/

/*
*	Allocates an object that has a pool
*	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

static	void	*AllocPool( NODE **pool, int objsize )
{
NODE	*optr;

	if ( *pool ) {
	
		optr  = *pool;
		*pool = optr->next;
TrackAlloc( optr->size );
	}
	else {
	
		optr = (NODE *) Malloc( objsize ) - 1;
	}
	
	optr->list = pool;
	optr->next = NULL;
	return( (void *) optr->data );
}
